"""
Test script for core exe_json
"""

# from __future__ import absolute_import, division, print_function

# from tests.util import unittest_reporter, glob_tests, cmp_dict

# import logging
# logger = logging.getLogger('exe')

# import os
# import sys
# import time
# import shutil
# import tempfile
# import random
# import string
# import subprocess
# from functools import partial

# try:
    # import cPickle as pickle
# except:
    # import pickle

# import unittest
# from unittest.mock import patch

# from tornado.testing import AsyncTestCase

# from iceprod.core import to_log,constants
# import iceprod.core.dataclasses
# import iceprod.core.functions
# import iceprod.core.exe
# import iceprod.core.exe_json
# from iceprod.core.jsonUtil import json_encode,json_decode,json_compressor


# class exe_json_test(AsyncTestCase):
    # def setUp(self):
        # super(exe_json_test,self).setUp()

        # curdir = os.getcwd()
        # self.test_dir = tempfile.mkdtemp(dir=curdir)
        # os.symlink(os.path.join(curdir, 'iceprod'),
                   # os.path.join(self.test_dir, 'iceprod'))
        # os.chdir(self.test_dir)
        # def cleanup():
            # os.chdir(curdir)
            # shutil.rmtree(self.test_dir)
        # self.addCleanup(cleanup)

        # # set offline mode
        # self.config = iceprod.core.exe.Config()
        # self.config.config['options']['offline'] = True

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # def test_01_ServerComms(self, Client):
        # """Test ServerComms"""
        # address = 'http://test'
        # passkey = 'ksdf8n4'
        # Client.return_value.request.return_value = 'e'

        # iceprod.core.exe_json.ServerComms(address, passkey, config=self.config)
        # self.assertTrue(Client.called)
        # logger.info('%r',Client.call_args[1])
        # self.assertEqual({'address':address,'token':passkey}, Client.call_args[1])

        # Client.reset_mock()
        # kwargs = {'ssl_cert':'cert','ssl_key':'key','cacert':'ca'}
        # iceprod.core.exe_json.ServerComms(address, passkey,
                                          # config=self.config, **kwargs)
        # self.assertTrue(Client.called)
        # expected = {'address':address}
        # expected.update(kwargs)
        # self.assertTrue(cmp_dict(expected, Client.call_args[1]))

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_10_download_task(self, Client):
        # """Test download_task"""
        # task = {'dataset':10}
        # c = Client.return_value
        # async def req(method, path, args=None):
            # if path.startswith('/task_actions/process'):
                # return {'task':'foo', 'task_id':'foo', 'job_id':'bar', 'dataset_id':'bar', 'task_index':1}
            # elif path.startswith('/config'):
                # return {'dataset_id':'bar'}
            # elif path.startswith('/jobs'):
                # return {'job_index':1}
            # elif path.startswith('/datasets'):
                # return {'dataset':1, 'jobs_submitted':10, 'tasks_submitted':1, 'debug':True}
        # c.request.side_effect = req

        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=self.config)
        # ret = (await rpc.download_task('gridspec'))[0]
        # self.assertIn('options', ret)
        # self.assertIn('task', ret['options'])
        # self.assertEqual(ret['options']['task'], 1)
        # self.assertIn('jobs_submitted', ret['options'])
        # self.assertEqual(ret['options']['jobs_submitted'], 10)

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_12_task_files(self, Client):
        # """Test task_files"""
        # c = Client.return_value
        # async def req(*args,**kwargs):
            # d = iceprod.core.dataclasses.Data()
            # d['remote'] = 'foo'
            # return {'files': [d]}
        # c.request.side_effect = req

        # self.config.config['options']['dataset_id'] = 'd'
        # self.config.config['options']['task_id'] = 't'

        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=self.config)
        # files = await rpc.task_files('d','t')

        # self.assertTrue(c.request.called)
        # logger.info(c.request.call_args[0])
        # logger.info('files: %r', files)
        # self.assertEqual(len(files), 1)
        # self.assertEqual(files[0]['remote'], 'foo')

        # async def req(*args,**kwargs):
            # return {'files': [{'movement':'blah'}]}
        # c.request.side_effect = req
        # with self.assertRaises(Exception):
            # await rpc.task_files('d','t')

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_15_processing(self, Client):
        # """Test processing"""
        # c = Client.return_value
        # async def req(*args,**kwargs):
            # return {}
        # c.request.side_effect = req

        # self.config.config['options']['task_id'] = 'task'

        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=self.config)
        # await rpc.processing('task')

        # self.assertTrue(c.request.called)
        # logger.info(c.request.call_args[0])
        # self.assertTrue({'status'}.issubset(
                        # c.request.call_args[0][-1]))

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_20_finish_task(self, Client):
        # """Test finish_task"""
        # c = Client.return_value
        # async def req(*args,**kwargs):
            # return {}
        # c.request.side_effect = req

        # self.config.config['options']['task_id'] = 'task'
        # stats = {'test':True}

        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=self.config)
        # await rpc.finish_task('task', stats=stats)

        # self.assertEqual(c.request.call_count, 2)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('task_stats', c.request.call_args_list[0][0][-1])

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_30_still_running(self, Client):
        # """Test still_running"""
        # c = Client.return_value
        
        # self.config.config['options']['task_id'] = 'task'

        # async def req(*args,**kwargs):
            # return {'status':'processing'}
        # c.request.side_effect = req
        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=self.config)
        # await rpc.still_running('task')
        # self.assertTrue(c.request.called)

        # async def req(*args,**kwargs):
            # return {'status':'reset'}
        # c.request.side_effect = req
        # with self.assertRaises(Exception):
            # await rpc.still_running('task')

        # c.request.side_effect = Exception('request error')
        # with self.assertRaises(Exception):
            # await rpc.still_running('task')

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_40_task_error(self, Client):
        # """Test task_error"""
        # c = Client.return_value
        # self.config.config['options']['task_id'] = 'task'
        
        # async def req(*args,**kwargs):
            # return {}
        # c.request.side_effect = req

        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=self.config)
        # await rpc.task_error('task')

        # self.assertEqual(c.request.call_count, 2)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('task_stats', c.request.call_args_list[0][0][-1])
        # self.assertIn('error_summary', c.request.call_args_list[0][0][-1])

        # c.request.reset_mock()
        # data = ''.join(random.choice(string.ascii_letters) for _ in range(10000))
        # await rpc.task_error('task', reason=data, start_time=time.time()-200)

        # self.assertEqual(c.request.call_count, 2)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('task_stats', c.request.call_args_list[0][0][-1])
        # self.assertIn('error_summary', c.request.call_args_list[0][0][-1])
        # self.assertEqual(data, c.request.call_args_list[0][0][-1]['error_summary'])
        # self.assertGreaterEqual(c.request.call_args_list[0][0][-1]['time_used'], 200)
        
        # c.request.side_effect = Exception('request error')
        # with self.assertRaises(Exception):
            # await rpc.task_error('task')

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_41_task_kill(self, Client):
        # """Test task_kill"""
        # c = Client.return_value
        # task_id = 'task'

        # async def req(*args,**kwargs):
            # return {}
        # c.request.side_effect = req

        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=None)
        # await rpc.task_kill(task_id)

        # self.assertEqual(c.request.call_count, 5)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('error_summary', c.request.call_args_list[0][0][-1])

        # c.request.reset_mock()
        # resources = {'cpu': 1, 'memory': 3.4, 'disk': 0.2}
        # await rpc.task_kill(task_id, resources=resources)
        
        # self.assertEqual(c.request.call_count, 5)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('error_summary', c.request.call_args_list[0][0][-1])
        # self.assertIn('resources', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['resources'], resources)

        # c.request.reset_mock()
        # resources = {'time': 34.2}
        # reason = 'testing'
        # await rpc.task_kill(task_id, resources=resources, reason=reason)
        
        # self.assertEqual(c.request.call_count, 5)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('error_summary', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['error_summary'], reason)
        # self.assertIn('resources', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['resources'], resources)

        # c.request.side_effect = Exception('request error')
        # with self.assertRaises(Exception):
            # await rpc.task_kill(task_id)

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_42_task_kill_sync(self, Client):
        # """Test task_kill"""
        # c = Client.return_value
        # task_id = 'task'

        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=None)
        # rpc.task_kill_sync(task_id)

        # self.assertEqual(c.request_seq.call_count, 5)
        # logger.info(c.request_seq.call_args_list[0][0])
        # self.assertIn('error_summary', c.request_seq.call_args_list[0][0][-1])

        # c.request_seq.reset_mock()
        # resources = {'cpu': 1, 'memory': 3.4, 'disk': 0.2}
        # rpc.task_kill_sync(task_id, resources=resources)
        
        # self.assertEqual(c.request_seq.call_count, 5)
        # logger.info(c.request_seq.call_args_list[0][0])
        # self.assertIn('error_summary', c.request_seq.call_args_list[0][0][-1])
        # self.assertIn('resources', c.request_seq.call_args_list[0][0][-1])
        # self.assertEqual(c.request_seq.call_args_list[0][0][-1]['resources'], resources)

        # c.request_seq.reset_mock()
        # resources = {'time': 34.2}
        # reason = 'testing'
        # rpc.task_kill_sync(task_id, resources=resources, reason=reason)
        
        # self.assertEqual(c.request_seq.call_count, 5)
        # logger.info(c.request_seq.call_args_list[0][0])
        # self.assertIn('error_summary', c.request_seq.call_args_list[0][0][-1])
        # self.assertEqual(c.request_seq.call_args_list[0][0][-1]['error_summary'], reason)
        # self.assertIn('resources', c.request_seq.call_args_list[0][0][-1])
        # self.assertEqual(c.request_seq.call_args_list[0][0][-1]['resources'], resources)

        # c.request_seq.side_effect = Exception('request error')
        # with self.assertRaises(Exception):
            # rpc.task_kill_sync(task_id)

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_50_uploadLogging(self, Client):
        # """Test uploading logfiles"""
        # c = Client.return_value
        # self.config.config['options']['task_id'] = 'task'

        # async def req(*args,**kwargs):
            # return {}
        # c.request.side_effect = req

        # data = ''.join([str(random.randint(0,10000)) for _ in range(100)])

        # filename = os.path.join(self.test_dir,str(random.randint(0,10000)))
        # with open(filename,'w') as f:
            # f.write(data)
        # name = 'testing'
        
        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=self.config)
        # await rpc._upload_logfile(name, filename)
        
        # self.assertTrue(c.request.called)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('name', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['name'], name)
        # self.assertIn('data', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['data'],
                         # data)

        # for f in constants.keys():
            # if f in ('stderr','stdout','stdlog'):
                # with open(constants[f],'w') as f:
                    # f.write(''.join([str(random.randint(0,10000))
                                     # for _ in range(100)]))

        # c.request.reset_mock()
        # await rpc.uploadLog()
        
        # self.assertTrue(c.request.called)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('name', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['name'], 'stdlog')
        # self.assertIn('data', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['data'],
                         # open(constants['stdlog']).read())

        # c.request.reset_mock()
        # await rpc.uploadErr()
        
        # self.assertTrue(c.request.called)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('name', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['name'], 'stderr')
        # self.assertIn('data', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['data'],
                         # open(constants['stderr']).read())

        # c.request.reset_mock()
        # await rpc.uploadOut()
        
        # self.assertTrue(c.request.called)
        # logger.info(c.request.call_args_list[0][0])
        # self.assertIn('name', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['name'], 'stdout')
        # self.assertIn('data', c.request.call_args_list[0][0][-1])
        # self.assertEqual(c.request.call_args_list[0][0][-1]['data'],
                        # open(constants['stdout']).read())


        # c.request.side_effect = Exception('request error')
        # with self.assertRaises(Exception):
            # await rpc._upload_logfile(name, filename)
        # with self.assertRaises(Exception):
            # await rpc.uploadLog()
        # with self.assertRaises(Exception):
            # await rpc.uploadErr()
        # with self.assertRaises(Exception):
            # await rpc.uploadOut()

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # async def test_60_update_pilot(self, Client):
        # """Test update_pilot"""
        # c = Client.return_value
        # pilot_id = 'pilot'
        # args = {'a': 1, 'b': 2}

        # async def req(*args,**kwargs):
            # return {}
        # c.request.side_effect = req
        
        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=None)
        # await rpc.update_pilot(pilot_id, **args)

        # self.assertTrue(c.request.called)
        # logger.info(c.request.call_args[0])
        # self.assertTrue({'a','b'}.issubset(c.request.call_args[0][-1]))

        # c.request.side_effect = Exception('request error')
        # with self.assertRaises(Exception):
            # await rpc.update_pilot(pilot_id, **args)

    # @patch('iceprod.core.exe_json.RestClient')
    # @unittest_reporter
    # def test_61_update_pilot_sync(self, Client):
        # """Test update_pilot"""
        # c = Client.return_value
        # pilot_id = 'pilot'
        # args = {'a': 1, 'b': 2}
        
        # rpc = iceprod.core.exe_json.ServerComms('a', 'p', config=None)
        # rpc.update_pilot_sync(pilot_id, **args)

        # self.assertTrue(c.request_seq.called)
        # logger.info(c.request_seq.call_args[0])
        # self.assertTrue({'a','b'}.issubset(c.request_seq.call_args[0][-1]))

        # c.request_seq.side_effect = Exception('request error')
        # with self.assertRaises(Exception):
            # rpc.update_pilot_sync(pilot_id, **args)

# def load_tests(loader, tests, pattern):
    # suite = unittest.TestSuite()
    # alltests = glob_tests(loader.getTestCaseNames(exe_json_test))
    # suite.addTests(loader.loadTestsFromNames(alltests,exe_json_test))
    # return suite



"""
Test script for daemon
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('daemon_test')

import os
import sys
import time
from functools import partial
import shutil
import tempfile
import signal
import stat
import multiprocessing

try:
    pass
except:
    pass

import unittest

from iceprod.server import daemon


def main(cfgfile,cfgdata):
    message_queue = multiprocessing.Queue()
    def handler2(signum, frame):
        logging.info('Signal handler2 called with signal %s' % signum)
        logging.info('Stopping...')
        message_queue.put('stop')
    def handler3(signum, frame):
        logging.info('Signal handler3 called with signal %s' % signum)
        logging.info('Killing...')
        message_queue.put('kill')
        time.sleep(2)
        sys.exit(1)
    signal.signal(signal.SIGINT, handler2)
    signal.signal(signal.SIGQUIT, handler3)
    with open('test','w') as f:
        f.write('test')
    while True:
        try:
            m = message_queue.get(True,10)
        except:
            raise
        else:
            if m == 'stop':
                break
            elif m == 'kill':
                break


class daemon_test(unittest.TestCase):
    def setUp(self):
        super(daemon_test,self).setUp()
        self.test_dir = tempfile.mkdtemp(dir=os.getcwd())

    def tearDown(self):
        shutil.rmtree(self.test_dir)
        super(daemon_test,self).tearDown()

    @unittest_reporter
    def test_01_Daemon(self):
        """Test daemon"""

        pidfile = os.path.expanduser(os.path.expandvars(
                        os.path.join(self.test_dir,'pidfile')))
        chdir = os.path.expanduser(os.path.expandvars(self.test_dir))
        umask = 0o077
        stdout = os.path.join(self.test_dir,'stdout')
        stderr = os.path.join(self.test_dir,'stderr')
        d = daemon.Daemon(pidfile,partial(main,'cfgfile','cfgdata'),
                   chdir=chdir,
                   umask=umask,
                   stdout=stdout,
                   stderr=stderr)
        multiprocessing.Process(target=d.start).start()
        time.sleep(1)
        if not os.path.exists(pidfile):
            raise Exception('pidfile creation failed')
        if not os.path.exists(os.path.join(chdir,'test')):
            raise Exception('chdir failed')
        st = os.stat(os.path.join(chdir,'test'))
        if oct(stat.S_IMODE(st[stat.ST_MODE])) != '0o600':
            logger.info('mode: %r',oct(stat.S_IMODE(st[stat.ST_MODE])))
            raise Exception('umask failed')

        d.stop()
        time.sleep(1)
        if os.path.exists(pidfile):
            raise Exception('pidfile still exists - stop failed')

        multiprocessing.Process(target=d.start).start()
        time.sleep(1)
        if not os.path.exists(pidfile):
            raise Exception('pidfile creation failed for start2')

        multiprocessing.Process(target=d.restart).start()
        time.sleep(2)
        if not os.path.exists(pidfile):
            raise Exception('pidfile creation failed for restart')

        d.kill()
        time.sleep(3)
        if os.path.exists(pidfile):
            raise Exception('pidfile still exists - kill failed')


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(daemon_test))
    suite.addTests(loader.loadTestsFromNames(alltests,daemon_test))
    return suite



import dataclasses

from wipac_dev_tools import from_environment_as_dataclass


@dataclasses.dataclass
class DefaultConfig:
    HOST : str = 'localhost'
    PORT : int = 8080
    DEBUG : bool = False
    OPENID_URL : str = ''
    OPENID_AUDIENCE : str = ''
    ICEPROD_WEB_URL : str = 'https://iceprod.icecube.aq'
    ICEPROD_API_ADDRESS : str = 'https://api.iceprod.icecube.aq'
    ICEPROD_API_CLIENT_ID : str = ''
    ICEPROD_API_CLIENT_SECRET : str = ''
    ICEPROD_CRED_ADDRESS : str = 'https://credentials.iceprod.icecube.aq'
    ICEPROD_CRED_CLIENT_ID : str = ''
    ICEPROD_CRED_CLIENT_SECRET : str = ''
    REDIS_HOST : str = 'localhost'
    REDIS_USER : str = ''
    REDIS_PASSWORD : str = ''
    REDIS_TLS : bool = False
    COOKIE_SECRET : str = ''
    PROMETHEUS_PORT: int = 0
    CI_TESTING: str = ''


def get_config() -> DefaultConfig:
    return from_environment_as_dataclass(DefaultConfig)



import time
import json
from unittest.mock import AsyncMock, MagicMock
import jwt
from pymongo import AsyncMongoClient
import pytest

import iceprod.credentials.service
import iceprod.credentials.util


GROUP = 'simprod'
USER = 'username'
DATASETS = ['dataset1', 'dataset2']
TASK_NAMES = ['alpha', 'bravo']


@pytest.mark.respx(using="httpx")
async def test_credentials_service_refresh_cred(respx_mock, monkeypatch):
    clients = json.dumps({
        'http://iceprod.test': {'client_id': 'id', 'client_secret': 'secret'}
    })
    rs = iceprod.credentials.service.RefreshService(MagicMock(), clients, 1, 1, 60)

    now = time.time()

    cred = {
        'url': 'http://iceprod.test',
        'type': 'oauth',
        'expiration': now,
        'last_use': now,
        'refresh_token': jwt.encode({'exp': now, 'iss': 'http://iceprod.test'}, 'secret'),
    }

    ref = jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test', 'refresh': 'refresh'}, 'secret')
    mock = respx_mock.post("http://iceprod.test/").respond(200, json={
        'access_token': jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test'}, 'secret'),
        'refresh_token': ref
    })

    authmock = MagicMock()
    authmock.return_value.token_url = 'http://iceprod.test'
    monkeypatch.setattr(iceprod.credentials.util, '_get_auth', authmock)

    new_cred = await rs.refresh_cred(cred)

    assert mock.called

    assert new_cred['refresh_token'] == ref
    assert new_cred['expiration'] == now+1000


async def test_credentials_service_refresh_empty(mongo_url, mongo_clear, respx_mock):
    db = AsyncMongoClient(mongo_url)['creds']
    clients = '{}'
    rs = iceprod.credentials.service.RefreshService(db, clients, 1, 1, 60)

    await rs._run_once()


@pytest.mark.respx(using="httpx", assert_all_called=False)
async def test_credentials_service_refresh_not_exp(mongo_url, mongo_clear, respx_mock):
    db = AsyncMongoClient(mongo_url)['creds']
    clients = '{}'
    rs = iceprod.credentials.service.RefreshService(db, clients, 1, 1, 60)

    now = time.time()

    await db.group_creds.insert_one({
        'groupname': GROUP,
        'url': 'http://iceprod.test',
        'type': 'oauth',
        'expiration': now+10000,
        'last_use': now,
        'refresh_token': jwt.encode({'exp': now, 'iss': 'http://iceprod.test'}, 'secret'),
    })

    mock = respx_mock.post("http://iceprod.test/").respond(200, json={})

    await rs._run_once()

    assert not mock.called
    
    assert rs.last_run_time is not None
    assert rs.last_success_time is not None


@pytest.mark.respx(using="httpx")
async def test_credentials_service_refresh_group(mongo_url, mongo_clear, respx_mock, monkeypatch):
    db = AsyncMongoClient(mongo_url)['creds']
    clients = json.dumps({
        'http://iceprod.test': {'client_id': 'id', 'client_secret': 'secret'}
    })
    rs = iceprod.credentials.service.RefreshService(db, clients, 1, 1, 60)

    now = time.time()

    await db.group_creds.insert_one({
        'groupname': GROUP,
        'url': 'http://iceprod.test',
        'type': 'oauth',
        'expiration': now,
        'last_use': now,
        'refresh_token': jwt.encode({'exp': now, 'iss': 'http://iceprod.test'}, 'secret'),
        'scope': '',
    })

    ref = jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test', 'refresh': 'refresh'}, 'secret')
    mock = respx_mock.post("http://iceprod.test/").respond(200, json={
        'access_token': jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test'}, 'secret'),
        'refresh_token': ref
    })

    authmock = MagicMock()
    authmock.return_value.token_url = 'http://iceprod.test'
    monkeypatch.setattr(iceprod.credentials.util, '_get_auth', authmock)

    await rs._run_once()

    assert mock.called

    ret = await db.group_creds.find_one({'url': 'http://iceprod.test'})
    assert ret
    assert ret['refresh_token'] == ref
    assert ret['expiration'] == now+1000


@pytest.mark.respx(using="httpx")
async def test_credentials_service_refresh_user(mongo_url, mongo_clear, respx_mock, monkeypatch):
    db = AsyncMongoClient(mongo_url)['creds']
    clients = json.dumps({
        'http://iceprod.test': {'client_id': 'id', 'client_secret': 'secret'}
    })
    rs = iceprod.credentials.service.RefreshService(db, clients, 1, 1, 60)

    now = time.time()

    await db.user_creds.insert_one({
        'username': USER,
        'url': 'http://iceprod.test',
        'type': 'oauth',
        'expiration': now,
        'last_use': now,
        'refresh_token': jwt.encode({'exp': now, 'iss': 'http://iceprod.test'}, 'secret'),
        'scope': '',
    })

    ref = jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test', 'refresh': 'refresh'}, 'secret')
    mock = respx_mock.post("http://iceprod.test/").respond(200, json={
        'access_token': jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test'}, 'secret'),
        'refresh_token': ref
    })

    authmock = MagicMock()
    authmock.return_value.token_url = 'http://iceprod.test'
    monkeypatch.setattr(iceprod.credentials.util, '_get_auth', authmock)

    await rs._run_once()

    assert authmock.called
    assert mock.called

    ret = await db.user_creds.find_one({'url': 'http://iceprod.test'})
    assert ret
    assert ret['refresh_token'] == ref
    assert ret['expiration'] == now+1000


@pytest.mark.respx(using="httpx")
async def test_credentials_service_refresh_dataset(mongo_url, mongo_clear, respx_mock, monkeypatch):
    db = AsyncMongoClient(mongo_url)['creds']
    clients = json.dumps({
        'http://iceprod.test': {'client_id': 'id', 'client_secret': 'secret'}
    })
    rs = iceprod.credentials.service.RefreshService(db, clients, 1, 1, 60)

    now = time.time()

    await db.dataset_creds.insert_one({
        'dataset_id': DATASETS[0],
        'task_name': TASK_NAMES[0],
        'url': 'http://iceprod.test',
        'type': 'oauth',
        'expiration': now,
        'last_use': now,
        'refresh_token': jwt.encode({'exp': now, 'iss': 'http://iceprod.test'}, 'secret'),
        'scope': '',
    })

    ref = jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test', 'refresh': 'refresh'}, 'secret')
    mock = respx_mock.post("http://iceprod.test/").respond(200, json={
        'access_token': jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test'}, 'secret'),
        'refresh_token': ref
    })

    authmock = MagicMock()
    authmock.return_value.token_url = 'http://iceprod.test'
    monkeypatch.setattr(iceprod.credentials.util, '_get_auth', authmock)

    await rs._run_once()

    assert authmock.called
    assert mock.called

    ret = await db.dataset_creds.find_one({'url': 'http://iceprod.test'})
    assert ret
    assert ret['refresh_token'] == ref
    assert ret['expiration'] == now+1000


@pytest.mark.respx(using="httpx")
async def test_credentials_service_exchange(respx_mock, monkeypatch):
    clients = json.dumps({
        'http://iceprod.test': {'client_id': 'id', 'client_secret': 'secret'}
    })
    rs = iceprod.credentials.service.RefreshService(AsyncMock(), clients, 1, 1, 60)

    now = time.time()

    cred = {
        'dataset_id': DATASETS[0],
        'task_name': TASK_NAMES[0],
        'url': 'http://iceprod.test',
        'type': 'oauth',
        'expiration': now,
        'last_use': now,
        'access_token': jwt.encode({'exp': now, 'iss': 'http://iceprod.test'}, 'secret'),
        'refresh_token': jwt.encode({'exp': now, 'iss': 'http://iceprod.test'}, 'secret'),
        'scope': '',
    }

    ref = jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test', 'refresh': 'refresh'}, 'secret')
    mock = respx_mock.post("http://iceprod.test/").respond(200, json={
        'access_token': jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test'}, 'secret'),
        'refresh_token': ref
    })

    authmock = MagicMock()
    authmock.return_value.token_url = 'http://iceprod.test'
    monkeypatch.setattr(iceprod.credentials.util, '_get_auth', authmock)

    ret = await rs.exchange_cred(cred, client_id='client_id')

    assert authmock.called
    assert mock.called

    assert ret
    assert ret['refresh_token'] == ref
    assert ret['expiration'] == now+1000


@pytest.mark.respx(using="httpx")
async def test_credentials_service_create(respx_mock, monkeypatch):
    clients = json.dumps({
        'http://iceprod.test': {'client_id': 'id', 'client_secret': 'secret', 'transfer_prefix': ['foo://']}
    })
    rs = iceprod.credentials.service.RefreshService(AsyncMock(), clients, 1, 1, 60)

    now = time.time()

    cred = {
        'url': 'http://iceprod.test',
        'transfer_prefix': 'foo://',
        'username': 'user',
        'scope': 'storage.read:/',
    }

    ref = jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test', 'refresh': 'refresh'}, 'secret')
    mock = respx_mock.post("http://iceprod.test/").respond(200, json={
        'access_token': jwt.encode({'exp': now+1000, 'iss': 'http://iceprod.test'}, 'secret'),
        'refresh_token': ref
    })

    authmock = MagicMock()
    authmock.return_value.token_url = 'http://iceprod.test'
    monkeypatch.setattr(iceprod.credentials.util, '_get_auth', authmock)

    ret = await rs.create_cred(**cred)

    assert authmock.called
    assert mock.called

    assert ret
    assert ret['refresh_token'] == ref
    assert ret['expiration'] == now+1000



from dataclasses import asdict, dataclass
import logging
from typing import Any

from tornado.web import HTTPError

from iceprod.common.mongo_queue import Message
from iceprod.core.config import Config as DatasetConfig, ValidationError
from iceprod.core.jsonUtil import json_decode, json_encode
from iceprod.core.parser import ExpParser
from iceprod.services.actions.submit import TokenSubmitter
from iceprod.services.base import AuthData, BaseAction


logger = logging.getLogger('edit_config')


@dataclass
class Fields:
    dataset_id: str
    config: str
    description: str | None = None


class Action(BaseAction):
    PRIORITY = 10

    async def create(self, args: dict[str, Any], *, auth_data: AuthData) -> str:
        try:
            data = Fields(**args)
        except Exception as e:
            raise HTTPError(400, reason=str(e))

        # check auth
        await self._manual_attr_auth('dataset_id', data.dataset_id, 'write', auth_data=auth_data)

        # validate config
        try:
            dc = DatasetConfig(json_decode(data.config))
            dc.fill_defaults()
            dc.validate()
            data.config = json_encode(dc.config)
        except ValidationError as e:
            raise HTTPError(400, reason=str(e))
        except Exception:
            raise HTTPError(400, reason='config is not valid')

        return await self._push(payload=asdict(data), priority=self.PRIORITY)

    async def run(self, message: Message) -> None:
        assert self._api_client and self._cred_client

        data = Fields(**message.payload)
        config = json_decode(data.config)

        dataset = await self._api_client.request('GET', f'/datasets/{data.dataset_id}')

        prev_config = await self._api_client.request('GET', f'/config/{data.dataset_id}')

        parser = ExpParser()
        config2 = config.copy()
        options = {
            'dataset': dataset['dataset'],
            'dataset_id': dataset['dataset_id'],
            'jobs_submitted': dataset['jobs_submitted']
        }
        options.update(config2['options'])
        config2['options'] = options

        # update tasks
        if len(config['tasks']) != len(prev_config['tasks']):
            raise Exception('cannot add/subtract tasks - create a new dataset')
        for task, prev_task in zip(config2['tasks'], prev_config['tasks']):
            config2['options']['task'] = task['name']
            if task['name'] != prev_task['name']:
                raise Exception('cannot edit task names - create a new dataset')
            if task['depends'] != prev_task['depends']:
                raise Exception('cannot edit task depends - create a new dataset')
            reqs = parser.parse(task['requirements'], env=config2)
            self._logger.info('new reqs: %r', reqs)
            if any('$' in val for val in reqs.values() if isinstance(val, str)):
                raise Exception('cannot update requirements - cannot parse expression')
            if reqs != prev_task['requirements']:
                await self._api_client.request('POST', f'/datasets/{data.dataset_id}/task_actions/bulk_requirements/{task["name"]}', reqs)

        # update config
        await self._api_client.request('PUT', f'/config/{data.dataset_id}', config)

        if data.description and data.description != dataset['description']:
            # update description
            args = {
                'description': data.description,
            }
            await self._api_client.request('PUT', f'/datasets/{data.dataset_id}/description', args)

        # check tokens
        ts = TokenSubmitter(
            logger=self._logger,
            cred_client=self._cred_client,
            jobs_submitted=dataset['jobs_submitted'],
            config=config,
            username=dataset['username'],
            group=dataset['group'],
        )
        if not await ts.tokens_exist(dataset_id=data.dataset_id):
            await ts.resubmit_tokens(dataset_id=data.dataset_id)

        self._logger.info('edit config complete')



"""
Buffer jobs and tasks into the queue.

Also known as "late materialization" of jobs and tasks,
this method finds datasets that are not fully materialized and
buffers a few more jobs and tasks into existence so they can be queued.
"""

import argparse
import asyncio
import logging

from rest_tools.client import RestClient

from iceprod.client_auth import add_auth_to_argparse, create_rest_client


logger = logging.getLogger('buffer_jobs_tasks')

MAX_PARALLEL_REQUESTS = 100


async def check_requests(requests: list[str], *, debug: bool, rest_client: RestClient) -> list[str]:
    logging.info('check_requests starting with %d requests', len(requests))
    futures = set()
    for mat_id in requests:
        futures.add(asyncio.create_task(rest_client.request('GET', f'/actions/materialization/{mat_id}')))

    requests2 = []
    while futures:
        logging.info('check_requests waiting on %d requests', len(futures))
        done, futures = await asyncio.wait(futures, timeout=1)
        for fut in done:
            try:
                ret = await fut
            except Exception as e:
                logger.warning('materialization request could not be queried')
                if debug:
                    raise e
            else:
                if ret['status'] in ('queued', 'processing'):
                    requests2.append(ret['id'])
                elif ret['status'] == 'complete':
                    logger.info(f'materialization request {ret["id"]} complete')
                elif ret['status'] == 'error':
                    logger.warning(f'materialization request {ret["id"]} failed')
                    if debug:
                        raise Exception(f'materialization failed: {ret}')
                else:
                    logger.warning(f'materialization request {ret["id"]} has unknown status')
                    if debug:
                        raise Exception(f'materialization failed: {ret}')
    return requests2


async def run(rest_client: RestClient, only_dataset: None | str = None, num: int = 1000, delay: int = 5, debug: bool = False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        only_dataset (str): dataset_id if we should only buffer a single dataset
        num (int): max number of jobs per dataset to buffer
        delay (float): wait N seconds between result checks
        debug (bool): debug flag to propagate exceptions
    """
    try:
        logger.info('starting materialization request')
        if only_dataset:
            datasets = [only_dataset]
        else:
            ret = await rest_client.request('GET', '/datasets', {'status': 'processing', 'keys': 'dataset_id'})
            datasets = list(ret.keys())

        futures = set()
        requests = []
        for i,dataset_id in enumerate(datasets):
            logger.info('starting request for dataset %s', dataset_id)
            futures.add(asyncio.create_task(rest_client.request('POST', '/actions/materialization', {'dataset_id': dataset_id, 'num': num})))
            while len(futures) > 8:
                logger.info('waiting for futures')
                done, futures = await asyncio.wait(futures, timeout=1)
                for fut in done:
                    try:
                        ret = await fut
                        materialization_id = ret['result']
                        logger.info(f'waiting for materialization request {materialization_id}')
                        requests.append(materialization_id)
                    except Exception as e:
                        logger.error('error creating materialization request for dataset', exc_info=True)
                        if debug:
                            raise e

            logger.warning('progress: %d of %d complete', i+1-len(futures)-len(requests), len(datasets))
            logger.info('%d futures outstanding', len(futures))
            logger.info('%d requests outstanding', len(requests))
            while len(requests) > MAX_PARALLEL_REQUESTS:
                requests = await check_requests(requests, debug=debug, rest_client=rest_client)
                await asyncio.sleep(delay)

        while futures:
            logger.info('waiting for futures')
            done, futures = await asyncio.wait(futures, timeout=1)
            for fut in done:
                try:
                    ret = await fut
                    materialization_id = ret['result']
                    logger.info(f'waiting for materialization request {materialization_id}')
                    requests.append(materialization_id)
                except Exception as e:
                    logger.error('error creating materialization request for dataset', exc_info=True)
                    if debug:
                        raise e

        while requests:
            logger.info('progress: %d of %d complete', len(datasets)-len(futures)-len(requests), len(datasets))
            logger.info('%d futures outstanding', len(futures))
            logger.info('%d requests outstanding', len(requests))
            requests = await check_requests(requests, debug=debug, rest_client=rest_client)
            await asyncio.sleep(delay)

        logger.info('complete')

    except Exception:
        logger.warning('materialization error', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('-d', '--dataset', type=str, default=None, help='dataset id (optional)')
    parser.add_argument('-n', '--num', type=int, default=100, help='number of jobs per dataset to buffer')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, only_dataset=args.dataset, num=args.num, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Make graphs (lifecycles, ...)
"""
from __future__ import absolute_import, division, print_function

import os
import sys
import json
from collections import OrderedDict

try:
    import pygraphviz as pgv
except ImportError:
    print('(optional) install pygraphviz to generate graphs')
    sys.exit(0)

from iceprod.server import get_pkgdata_filename

def main():
    table_filename = get_pkgdata_filename('iceprod.server','data/etc/db_config.json')
    db_tables = json.load(open(table_filename),object_pairs_hook=OrderedDict)

    for k in db_tables['status_graphs']:
        outfile_name = os.path.join('static','lifecycle_'+k+'.png')
        if os.path.exists(outfile_name) and os.path.getmtime(outfile_name) > os.path.getmtime(table_filename):
            print('graph',outfile_name,'already exists. skipping')
            continue

        G = pgv.AGraph(strict=False,directed=True)
        G.add_nodes_from(db_tables['status_options'][k])
        for row in db_tables['status_graphs'][k]:
            if row[-1] == 'std':
                c = 'cornflowerblue'
            elif row[-1] == 'auto':
                c = 'cyan2'
            elif row[-1] == 'debug':
                c = 'chartreuse2'
            elif row[-1] == 'manual':
                c = 'firebrick2'
            G.add_edge(row[0],row[1],color=c)
        G.draw(outfile_name, prog='dot')

if __name__ == '__main__':
    main()



# make dataclasses.js from dataclasses.py
import inspect
import json
import os
import sys
current_path = os.path.abspath(os.path.dirname(__file__))
sys.path.append(current_path)
from iceprod.core import dataclasses
dcs = {}
names = dataclasses._plurals.copy()
for name, obj in inspect.getmembers(dataclasses,inspect.isclass):
    if name[0] != '_' and dict in inspect.getmro(obj):
        dcs[name] = obj().output()
        names[name] = obj.plural
data = {'classes':dcs,'names':names}
with open(os.path.join(current_path,'iceprod','server','data','www','dataclasses.js'),'w') as f:
    f.write('var dataclasses='+json.dumps(data,separators=(',',':'))+';')



import re
from unittest.mock import AsyncMock, MagicMock

import pytest
import requests
from rest_tools.client import RestClient

from iceprod.services.actions.materialization.materialize import Materialize


async def test_materialization_server_request(server):
    client = server(roles=['system'])
    args = {
        'dataset_id': 'd123',
    }
    ret = await client.request('POST', '/actions/materialization', args)
    assert 'result' in ret

    ret = await client.request('GET', f'/actions/materialization/{ret["result"]}')
    assert ret['status'] == 'queued'


async def test_materialization_server_request_dataset(server, requests_mock):
    requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)
    requests_mock.register_uri('POST', re.compile('localhost'), real_http=True)

    requests_mock.post('http://test.iceprod/auths', status_code=200, json={})

    # test system auth
    client = server(roles=['system'])
    ret = await client.request('POST', '/actions/materialization/datasets/d123', {})
    mat_id = ret['result']
    ret = await client.request('POST', '/actions/materialization/datasets/d123', {})
    mat_id2 = ret['result']

    assert mat_id == mat_id2

    # test user auth
    client = server(roles=['user'])
    ret = await client.request('POST', '/actions/materialization/datasets/d123', {})
    mat_id = ret['result']
    ret = await client.request('POST', '/actions/materialization/datasets/d123', {})
    mat_id2 = ret['result']

    assert mat_id == mat_id2

    ret = await client.request('GET', '/actions/materialization/datasets/d123')
    assert ret['status'] == 'queued'
    assert ret['id'] == mat_id

    # test no auth
    requests_mock.post('http://test.iceprod/auths', status_code=403, json={})
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/actions/materialization/datasets/d123')
    assert exc_info.value.response.status_code == 403



def test_materialize_init():
    Materialize(MagicMock())

async def test_materialize_run_none(requests_mock):
    rc = RestClient('http://test.iceprod')
    m = Materialize(rc)
    m.buffer_job = AsyncMock(return_value=1)

    requests_mock.get('http://test.iceprod/dataset_summaries/status', json={})

    ret = await m.run_once(num=1)
    assert ret is True
    assert m.buffer_job.call_count == 0

async def test_materialize_run_one_job(requests_mock):
    rc = RestClient('http://test.iceprod')
    m = Materialize(rc)
    m.buffer_job = AsyncMock(return_value=1)

    requests_mock.get('http://test.iceprod/dataset_summaries/status', json={
        'processing': ['did123']
    })

    requests_mock.get('http://test.iceprod/datasets/did123', json={
        'dataset_id': 'did123',
        'dataset': 123,
        'status': 'processing',
        'tasks_per_job': 1,
        'jobs_submitted': 10,
        'tasks_submitted': 10,
        'debug': False,
    })

    requests_mock.get('http://test.iceprod/datasets/did123/job_counts/status', json={})
    requests_mock.get('http://test.iceprod/datasets/did123/task_counts/status', json={})
    requests_mock.get('http://test.iceprod/datasets/did123/jobs', json={})

    await m.run_once(num=1)

    m.buffer_job.assert_called_once()

async def test_materialize_buffer_job_no_depends(requests_mock):
    rc = RestClient('http://test.iceprod')
    m = Materialize(rc)
    config = {
        'tasks': [
            {
                'name': 'foo',
            }
        ],
        'options': {}
    }
    m.get_config = AsyncMock(return_value=config)
    m.prio = MagicMock()
    m.prio.get_task_prio = AsyncMock(return_value=1.)

    dataset = {
        'dataset_id': 'did123',
        'dataset': 123,
        'status': 'processing',
        'tasks_per_job': 1,
        'jobs_submitted': 10,
        'tasks_submitted': 10,
        'debug': False,
    }

    requests_mock.post('http://test.iceprod/jobs', json={'result': 'j123'})
    requests_mock.post('http://test.iceprod/tasks', json={'result': 't123'})
    requests_mock.patch('http://test.iceprod/tasks/t123', json={})

    ret = await m.buffer_job(dataset, 0)

    assert ret == 1


async def test_materialize_buffer_job_incomplete(monkeypatch, requests_mock):
    rc = RestClient('http://test.iceprod')
    m = Materialize(rc)
    config = {
        'tasks': [
            {
                'name': 'foo',
            },
            {
                'name': 'bar',
            },
            {
                'name': 'baz',
            }
        ],
        'options': {}
    }
    m.get_config = AsyncMock(return_value=config)
    prio_mock = MagicMock()
    monkeypatch.setattr('iceprod.services.actions.materialization.materialize.Priority', prio_mock)
    prio_mock.return_value.get_task_prio = AsyncMock(return_value=1.)

    requests_mock.get('http://test.iceprod/datasets/did123', json={
        'dataset_id': 'did123',
        'dataset': 123,
        'status': 'processing',
        'tasks_per_job': 3,
        'jobs_submitted': 10,
        'tasks_submitted': 10,
        'debug': False,
    })
    
    requests_mock.get('http://test.iceprod/datasets/did123/job_counts/status', json={
        'processing': 1,
    })
    requests_mock.get('http://test.iceprod/datasets/did123/task_counts/status', json={
        'idle': 1,
    })
    requests_mock.get('http://test.iceprod/datasets/did123/jobs', json={
        'j123': {
            'job_id': 'j123',
            'job_index': 0,
        },
    })
    requests_mock.get('http://test.iceprod/datasets/did123/tasks', json={
        't0': {
            'task_id': 't0',
            'job_id': 'j123',
            'task_index': 0,
        },
    })

    requests_mock.post('http://test.iceprod/jobs', json={'result': 'j2'})
    requests_mock.post('http://test.iceprod/tasks', json={'result': 't0'})
    requests_mock.patch('http://test.iceprod/tasks/t0', json={})

    ret = await m.run_once(only_dataset='did123', num=0)

    calls = [h for h in requests_mock.request_history if h.url == 'http://test.iceprod/tasks']
    assert len(calls) == 2


async def test_materialize_buffer_job_no_tasks(monkeypatch, requests_mock):
    rc = RestClient('http://test.iceprod')
    m = Materialize(rc)
    config = {
        'tasks': [
            {
                'name': 'foo',
            },
            {
                'name': 'bar',
            },
            {
                'name': 'baz',
            }
        ],
        'options': {}
    }
    m.get_config = AsyncMock(return_value=config)
    prio_mock = MagicMock()
    monkeypatch.setattr('iceprod.services.actions.materialization.materialize.Priority', prio_mock)
    prio_mock.return_value.get_task_prio = AsyncMock(return_value=1.)

    requests_mock.get('http://test.iceprod/datasets/did123', json={
        'dataset_id': 'did123',
        'dataset': 123,
        'status': 'processing',
        'tasks_per_job': 3,
        'jobs_submitted': 10,
        'tasks_submitted': 10,
        'debug': False,
    })
    
    requests_mock.get('http://test.iceprod/datasets/did123/job_counts/status', json={
        'processing': 1,
    })
    requests_mock.get('http://test.iceprod/datasets/did123/task_counts/status', json={
    })
    requests_mock.get('http://test.iceprod/datasets/did123/jobs', json={
        'j123': {
            'job_id': 'j123',
            'job_index': 0,
        },
    })
    requests_mock.get('http://test.iceprod/datasets/did123/tasks', json={
    })

    requests_mock.post('http://test.iceprod/jobs', json={'result': 'j2'})
    requests_mock.post('http://test.iceprod/tasks', json={'result': 't0'})
    requests_mock.patch('http://test.iceprod/tasks/t0', json={})

    ret = await m.run_once(only_dataset='did123', num=0)

    calls = [h for h in requests_mock.request_history if h.url == 'http://test.iceprod/tasks']
    assert len(calls) == 3



"""
Interface for configuring and submitting jobs on a computing cluster.
Do not use this class directly. Instead use one of the batch plugins
that inherit from this class.
"""

import asyncio
from collections import Counter
from enum import StrEnum
import json
import os
import logging
from copy import deepcopy
from pathlib import Path
import socket
from typing import Any, Protocol

from asyncache import cached, cachedmethod  # type: ignore
from cachetools import TTLCache
from cachetools.func import ttl_cache
from prometheus_client import Info
import requests.exceptions
from wipac_dev_tools.prometheus_tools import GlobalLabels, AsyncPromWrapper, AsyncPromTimer

from iceprod.server.priority import Priority
from iceprod.util import VERSION_STRING
from iceprod.core import functions
from iceprod.core.config import Task, Job, Dataset
from iceprod.core.defaults import add_default_options
from iceprod.core.resources import Resources
from iceprod.common.prom_utils import HistogramBuckets
from iceprod.server.states import JOB_STATUS_START
from iceprod.server.util import nowstr


logger = logging.getLogger('grid')


@ttl_cache(ttl=3600)
def get_host():
    """Cache the host fqdn for 1 hour"""
    return socket.getfqdn()


class GridTask(Protocol):
    """Protocol for grid task dataclass"""
    dataset_id: str | None
    task_id: str | None
    instance_id: str | None


class GridStatus(StrEnum):
    QUEUED = 'queued'
    PROCESSING = 'processing'


class BaseGrid:
    """
    Interface for a generic job distribution system.
    Do not use this class directly.  Use one of the plugins.
    """

    def __init__(self, cfg, rest_client, cred_client):
        self.cfg = cfg
        self.rest_client = rest_client
        self.cred_client = cred_client

        queue_cfg = self.cfg['queue']

        # site name
        self.site = ''
        if 'site' in queue_cfg:
            self.site = queue_cfg['site']

        # task queue params
        self.site_requirements = queue_cfg['resources']
        self.site_query_params = {}
        if self.site:
            self.site_requirements['site'] = self.site
            if queue_cfg.get('exclusive', False):
                logger.info('site requirements: must match site name')
                self.site_query_params['requirements.site'] = self.site

            if 'gpu' in self.site.lower():
                logger.info('site requirements: GPU site!')
                self.site_requirements['gpu'] = 1

        # directories
        self.credentials_dir = Path(os.path.expanduser(os.path.expandvars(
            queue_cfg['credentials_dir'])))
        self.credentials_dir.mkdir(mode=0o700, parents=True, exist_ok=True)

        self.submit_dir = Path(os.path.expanduser(os.path.expandvars(
            queue_cfg['submit_dir'])))
        self.submit_dir.mkdir(mode=0o700, parents=True, exist_ok=True)

        # dataset lookup cache
        self.dataset_cache = TTLCache(maxsize=100, ttl=60)

        i = Info('iceprod', 'IceProd information')
        i.info({
            'name': str(self.site),
            'type': 'grid',
            'queue_type': queue_cfg.get('type',''),
            'version': VERSION_STRING,
            'exclusive': str(queue_cfg.get('exclusive', False)),
        })

        self.prometheus = GlobalLabels({
            'site': str(self.site),
            'type': queue_cfg.get('type',''),
            'gpu': str(self.site_requirements.get('gpu', 0)),
        })

    # Functions to override #

    async def run(self):
        """Override the `ActiveJobs` and `JobActions` batch submission / monitoring classes"""
        raise NotImplementedError()

    def queue_dataset_status(self) -> dict[GridStatus, Counter[str]]:
        """Get the current queue job counts by dataset and job status."""
        raise NotImplementedError()

    # Common functions #

    @cachedmethod(lambda self: self.dataset_cache)
    @AsyncPromTimer(lambda self: self.prometheus.histogram('iceprod_grid_dataset_lookup', 'IceProd grid.dataset_lookup calls', buckets=HistogramBuckets.SECOND))
    async def dataset_lookup(self, dataset_id: str) -> Dataset:
        """
        Lookup a dataset fromn the REST API.

        Uses caching with TTL 60 seconds.

        Args:
            dataset_id: dataset id

        Returns:
            iceprod.core.config.Dataset object
        """
        ret = await Dataset.load_from_api(dataset_id, rest_client=self.rest_client)
        ret.fill_defaults()
        ret.validate()
        return ret

    @cached(TTLCache(10, 900), key=lambda _: 'self')
    async def _get_priority_object(self) -> Priority:
        p = Priority(rest_client=self.rest_client)
        await p._populate_dataset_cache()
        return p

    @AsyncPromWrapper(lambda self: self.prometheus.counter('iceprod_grid_queue_tasks', 'IceProd grid tasks queued', labels=['step'], finalize=False))
    @AsyncPromTimer(lambda self: self.prometheus.histogram('iceprod_grid_tasks_to_queue', 'IceProd grid.tasks_to_queue calls'))
    async def get_tasks_to_queue(self, prom_counter, num: int) -> list[Task]:
        """
        Get new tasks to queue from the REST API.

        Args:
            num: number of tasks to retrieve

        Returns:
            list of iceprod.core.config.Task objects
        """
        # get current count of queued jobs (ignore processing)
        cur_jobs: Counter[str] = Counter(self.queue_dataset_status().get(GridStatus.QUEUED, {}))
        priorities = await self._get_priority_object()
        dataset_prios = {}
        for dataset_id in cur_jobs:
            dataset_prios[dataset_id] = await priorities.get_dataset_prio(dataset_id)

        # order the ignore list by number of queued jobs / dataset priority
        # this should slightly favor higher priority datasets, but allow lower priority datasets to come in
        ignore_datasets_list = sorted(cur_jobs, key=lambda x: cur_jobs[x]/dataset_prios[x] if dataset_prios.get(x,None) else 1000000, reverse=True)
        logger.info('ignore_dataset_list=%r', ignore_datasets_list)

        # get tasks to run from REST API, and convert to batch jobs
        try:
            ret = await self.rest_client.request('POST', '/task_actions/queue_many', {
                'num': num,
                'requirements': self.site_requirements,
                'dataset_deprio': ignore_datasets_list,
                'query_params': self.site_query_params,
            })
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return []
            raise

        futures = set()
        for row in ret:
            futures.add(asyncio.create_task(self._convert_to_task(row)))

        tasks_queued = len(ret)
        prom_counter.labels({'step': 'raw'}).inc(tasks_queued)
        logging.info('got %d tasks to queue', tasks_queued)

        tasks = []
        for f in asyncio.as_completed(futures):
            try:
                task = await f
            except Exception:  # already logged in function
                continue
            try:
                # add default resource requirements
                task.requirements = self._get_resources(task)
            except Exception:
                logger.warning('cannot get task resources for %s.%s', task.dataset.dataset_id, task.task_id, exc_info=True)
                continue
            try:
                # add oauth tokens
                task.oauth_tokens = await self._get_dataset_credentials(task)
            except Exception:
                logger.warning('cannot get oauth tokens for %s.%s', task.dataset.dataset_id, task.task_id, exc_info=True)
                continue
            tasks.append(task)

        tasks_queued2 = len(tasks)
        prom_counter.labels({'step': 'processed'}).inc(tasks_queued2)
        logging.info('got %d Task objects to queue', tasks_queued2)

        return tasks

    async def _convert_to_task(self, task):
        """Convert from basic task dict to a Task object"""
        try:
            d = deepcopy(await self.dataset_lookup(task['dataset_id']))
            # don't bother looking up the job status - trust that if we got a task, we're in processing
            j = Job(dataset=d, job_id=task['job_id'], job_index=task['job_index'], status=JOB_STATUS_START)
            t = Task(
                dataset=d,
                job=j,
                task_id=task['task_id'],
                task_index=task['task_index'],
                instance_id=task['instance_id'],
                name=task['name'],
                depends=task['depends'],
                requirements=task['requirements'],
                status=task['status'],
                site=self.site,
                stats={},
                task_files=[],
            )
            await t.load_task_files_from_api(self.rest_client)

            # load some config defaults
            config = t.dataset.config
            if (not config['options'].get('site_temp','')) and self.cfg['queue'].get('site_temp', ''):
                config['options']['site_temp'] = self.cfg['queue']['site_temp']
            add_default_options(config['options'])

        except Exception:
            logger.warning('Error converting task dict to task: %s.%s', task['dataset_id'], task['task_id'], exc_info=True)
            raise

        return t

    @cached(TTLCache(1024, 60), key=lambda _,t: (t.dataset.dataset_id, t.name))
    async def _get_dataset_credentials(self, task: Task) -> list[Any]:
        # todo: handle non-oauth credentials, like s3
        # todo: delayed get to see if we already have these tokens in condor
        ret = []
        if client_id := self.cfg['oauth_condor_client_id']:
            # do dataset creds, with downscoping
            for prefix, scope in task.get_task_config().get('token_scopes', {}).items():
                args = {'client_id': client_id, 'transfer_prefix': prefix, 'new_scope': scope}
                ret2 = await self.cred_client.request('GET', f'/datasets/{task.dataset.dataset_id}/exchange', args)
                ret.extend(ret2)
            logging.info('dataset %s creds with downscoping: %d', task.dataset.dataset_id, len(ret))
            # do dataset-task creds
            args = {'client_id': client_id}
            ret2 = await self.cred_client.request('GET', f'/datasets/{task.dataset.dataset_id}/tasks/{task.name}/exchange', args)
            logging.info('dataset %s task creds: %d', task.dataset.dataset_id, len(ret2))
            ret.extend(ret2)
        return ret

    @cached(TTLCache(10, 60), key=lambda _: 'self')
    async def get_scratch_credentials(self):
        if client_id := self.cfg['oauth_condor_client_id']:
            args = {'client_id': client_id, 'transfer_prefix': self.cfg['queue']['site_temp']}
            ret = await self.cred_client.request('GET', '/users/ice3simusr/exchange', args)
            if ret:
                with open(self.credentials_dir / 'scratch', 'w') as f:
                    json.dump(ret, f)

    @staticmethod
    def _get_resources(task):
        """
        Combine task resources with default resources.

        Args:
            task (Task): IceProd Task
        Returns:
            dict: resources
        """
        resource = deepcopy(Resources.defaults)
        for k in resource:
            if isinstance(resource[k], list):
                resource[k] = len(resource[k])  # type: ignore
        values = {}
        try:
            for k in task.requirements:
                if k in resource and task.requirements[k]:
                    try:
                        if isinstance(resource[k], int):
                            values[k] = int(task.requirements[k])
                        elif isinstance(resource[k], float):
                            values[k] = float(task.requirements[k])
                        else:
                            values[k] = task.requirements[k]
                    except Exception:
                        logger.warning('bad reqs value for task %r', task)
                elif k == 'os' and task.requirements[k]:
                    logger.debug('OS req: %s', task.requirements[k])
                    values['os'] = task.requirements[k] if isinstance(task.requirements[k], list) else [task.requirements[k]]
        except TypeError:
            logger.warning('task.requirements: %r', task.requirements)
            raise
        resource.update(values)
        return resource

    @AsyncPromWrapper(lambda self: self.prometheus.gauge('iceprod_grid_tasks', 'IceProd grid tasks IceProd thinks are on the queue'))
    async def get_tasks_on_queue(self, prom_gauge) -> list:
        """
        Get all tasks that are "assigned" to this queue.

        Returns:
            list of tasks
        """
        args = {
            'status': 'queued|processing',
            'site': self.site,
            'keys': 'dataset_id|task_id|instance_id|status|status_changed',
        }
        try:
            tasks = await self.rest_client.request('GET', '/tasks', args)
            ret = tasks['tasks']
            prom_gauge.set(len(ret))
            return ret
        except requests.exceptions.HTTPError:
            logger.warning('cannot get tasks on queue', exc_info=True)
            return []

    # Task Actions #

    async def _upload_log(self, task: GridTask, name: str, data: str):
        """
        Upload a log to the IceProd API.

        Args:
            task: IceProd task info
            name: log name
            data: log text data
        """
        args = {
            'dataset_id': task.dataset_id,
            'task_id': task.task_id,
            'name': name,
            'data': data,
        }
        try:
            await self.rest_client.request('POST', '/logs', args)
        except requests.exceptions.HTTPError:
            logger.warning('cannot upload log', exc_info=True)

    async def _upload_stats(self, task: GridTask, stats: dict):
        """
        Upload task statistics to the IceProd API.

        Args:
            task: IceProd task info
            stats: stats dict
        """
        stats_cp = stats.copy()

        hostname = functions.gethostname()
        domain = '.'.join(hostname.split('.')[-2:])
        site = self.site
        if 'site' in stats_cp:
            site = stats_cp.pop('site', '')
        args = {
            'dataset_id': task.dataset_id,
            'hostname': hostname,
            'domain': domain,
            'site': site,
            'resources': stats_cp.pop('resources', {}),
            'task_stats': stats_cp,
            'time': nowstr(),
        }
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_stats', args)
        except requests.exceptions.HTTPError:
            logger.warning('cannot upload stats', exc_info=True)

    @AsyncPromTimer(lambda self: self.prometheus.histogram('iceprod_grid_task_idle', 'IceProd grid.task_idle calls', buckets=HistogramBuckets.MINUTE))
    async def task_idle(self, task: GridTask):
        """
        Tell IceProd API a task is now idle on the queue (put back to "queue" status).

        Args:
            task: IceProd task info
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args = {
            'status': 'queued',
            'instance_id': task.instance_id,
        }
        try:
            await self.rest_client.request('PUT', f'/tasks/{task.task_id}/status', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise

    @AsyncPromTimer(lambda self: self.prometheus.histogram('iceprod_grid_task_processing', 'IceProd grid.task_processing calls', buckets=HistogramBuckets.MINUTE))
    async def task_processing(self, task: GridTask, site: str | None = None):
        """
        Tell IceProd API a task is now processing (put in the "processing" status).

        Args:
            task: IceProd task info
            site: computing site the task is running at
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args = {
            'instance_id': task.instance_id,
        }
        if site:
            args['site'] = site
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/processing', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise

    @AsyncPromTimer(lambda self: self.prometheus.histogram('iceprod_grid_task_reset', 'IceProd grid.task_reset calls', buckets=HistogramBuckets.MINUTE))
    async def task_reset(self, task: GridTask, reason: str | None = None, stats: dict | None = None):
        """
        Tell IceProd API a task should be reset back to the "waiting" status.

        Args:
            task: IceProd task info
            stats: task resource statistics
            reason: A reason for failure
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args = {
            'instance_id': task.instance_id,
        }
        if stats:
            args['resources'] = stats.get('resources', {})
            if site := stats.get('site'):
                args['site'] = site
        if reason:
            args['reason'] = reason
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/reset', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise
        else:
            if reason:
                await self._upload_log(task, 'stdlog', reason)

    @AsyncPromTimer(lambda self: self.prometheus.histogram('iceprod_grid_task_failure', 'IceProd grid.task_failure calls', buckets=HistogramBuckets.MINUTE))
    async def task_failure(self, task: GridTask, reason: str | None = None, stats: dict | None = None, stdout: Path | None = None, stderr: Path | None = None):
        """
        Tell IceProd API a task should be put in the "failed" status.

        Stats format:
            {
                resources: {cpu: 1, gpu: 1, ...},
                ...
            }

        Args:
            task: IceProd task info
            reason: A reason for failure
            stats: task resource statistics
            stdout: path to stdout file
            stderr: path to stderr file
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        args: dict[str, Any] = {
            'instance_id': task.instance_id,
        }
        if reason:
            args['reason'] = reason
        if stats:
            args['resources'] = stats.get('resources', {})
            if site := stats.get('site'):
                args['site'] = site
        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/failed', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise
        else:
            if stats:
                await self._upload_stats(task, stats)
            if stdout and stdout.exists():
                await self._upload_log(task, 'stdout', stdout.read_text())
            if stderr and stderr.exists():
                await self._upload_log(task, 'stderr', stderr.read_text())
            if reason:
                await self._upload_log(task, 'stdlog', reason)

    @AsyncPromTimer(lambda self: self.prometheus.histogram('iceprod_grid_task_success', 'IceProd grid.task_success calls', buckets=HistogramBuckets.MINUTE))
    async def task_success(self, task: GridTask, stats: dict | None = None, stdout: Path | None = None, stderr: Path | None = None):
        """
        Tell IceProd API a task was successfully completed.

        Args:
            task: IceProd task info
            stats: task resource statistics
            stdout: path to stdout file
            stderr: path to stderr file
        """
        if not task.task_id or not task.instance_id:
            raise RuntimeError("Either task_id or instance_id is empty")

        site = self.site
        if stats and 'site' in stats:
            site = stats['site']
        args = {
            'site': site,
            'instance_id': task.instance_id,
        }
        if stats:
            resources = stats.get('resources', {})
            if site := stats.get('site'):
                args['site'] = site
            if 'time' in resources:
                args['time_used'] = resources['time']*3600.

        try:
            await self.rest_client.request('POST', f'/tasks/{task.task_id}/task_actions/complete', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404:
                raise
        else:
            if stats:
                await self._upload_stats(task, stats)
            if stdout and stdout.exists():
                await self._upload_log(task, 'stdout', stdout.read_text())
            if stderr and stderr.exists():
                await self._upload_log(task, 'stderr', stderr.read_text())




from dataclasses import dataclass
from functools import cached_property
import inspect
from logging import Logger
import logging
from typing import Any

import requests
from rest_tools.client import RestClient
from tornado.web import HTTPError

from iceprod.common.mongo_queue import AsyncMongoQueue, Message, Payload
from iceprod.rest.base_handler import APIBase


type HandlerTypes = list[tuple[str, Any, dict[str, Any]]]


class TimeoutException(Exception):
    def __init__(self, *args, update_payload: dict[str, Any] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_payload = update_payload


@dataclass(frozen=True)
class AuthData:
    username: str
    groups: list[str]
    roles: list[str]
    token: dict[str, Any]


async def check_attr_auth(arg: str, val: str, role: str, *, auth_data: AuthData, rest_client: RestClient, token_role_bypass: list[str] = ['admin', 'system']):
    """
    Manually run check_attr_auth and raise an error if we fail the auth check.

    Args:
        arg: attribute name to check
        val: attribute value
        role: the role to check for (read|write)
        auth_data: request auth data
        token_role_bypass: token roles that bypass this auth (default: admin,system)

    Raises:
        HTTPError: when not authorized
    """
    if any(r in auth_data.roles for r in token_role_bypass):
        logging.debug('token role bypass')
        return True
    args = {
        'name': arg,
        'value': val,
        'role': role,
        'username': auth_data.username,
        'groups': auth_data.groups,
    }
    try:
        await rest_client.request('POST', '/auths', args)
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 403:
            raise HTTPError(403, reason='auth failed')
        else:
            raise HTTPError(500, reason='auth could not be completed')


class BaseAction:
    def __init__(self, queue: AsyncMongoQueue, logger: Logger, api_client: RestClient | None = None, cred_client: RestClient | None = None):
        self._queue = queue
        self._logger = logger
        self._api_client = api_client
        self._cred_client = cred_client

    @cached_property
    def action_type(self) -> str:
        mod = inspect.getmodule(self.__class__)
        if not mod or not mod.__package__:
            raise Exception('cannot get action type')
        if mod.__package__ == 'iceprod.services.actions':
            return mod.__name__.rsplit('.', 1)[1]
        else:
            return mod.__package__.rsplit('.', 1)[1]

    def extra_handlers(self) -> HandlerTypes:
        return []

    async def create(self, args: dict[str, Any], *, auth_data: AuthData) -> str:
        raise NotImplementedError()

    async def _push(self, *, payload: Payload, filter_payload: Payload | None = None, priority: int = 0) -> str:
        payload['type'] = self.action_type
        if not filter_payload:
            return await self._queue.push(payload=payload, priority=priority)
        else:
            return await self._queue.push_if_not_exists(payload=payload, filter_payload=filter_payload, priority=priority)

    async def run(self, message: Message) -> None:
        raise NotImplementedError()

    async def _manual_attr_auth(self, arg: str, val: str, role: str, *, auth_data: AuthData):
        """
        Manually run check_attr_auth and return a boolean.

        Args:
            arg: attribute name to check
            val: attribute value
            role: the role to check for (read|write)
            auth_data: request auth data
            token_role_bypass: token roles that bypass this auth (default: admin,system)

        Raises:
            HTTPError: when not authorized
        """
        assert self._api_client
        await check_attr_auth(arg, val, role, auth_data=auth_data, rest_client=self._api_client)


class BaseHandler(APIBase):
    def initialize(self, *args, message_queue: AsyncMongoQueue, rest_client: RestClient, action: BaseAction | None = None, **kwargs):  # type: ignore[override]
        super().initialize(*args, **kwargs)
        self.message_queue = message_queue
        self.rest_client = rest_client
        self.action = action

    async def check_attr_auth(self, arg: str, val: str, role: str):
        auth = AuthData(
            username=self.current_user,
            groups=self.auth_groups,
            roles=self.auth_roles,
            token=self.auth_data,
        )
        await check_attr_auth(arg, val, role, auth_data=auth, rest_client=self.rest_client)



from copy import deepcopy
from datetime import datetime
import logging
import re
from typing import Any
from unittest.mock import AsyncMock, MagicMock
from uuid import uuid4

import pytest

from iceprod.common.mongo_queue import Message
from iceprod.core.config import Config
from iceprod.core.jsonUtil import json_encode
import iceprod.services.actions.submit
from iceprod.services.actions.edit_config import Action


@pytest.mark.parametrize('config,error', [
    ({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }],
        'version': 3.2,
    }, None),
    ({'tasks':[
        {
            'name': 'testing',
            'trays': []
        }
    ]}, 'Validation error'),
    ({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': '',
                    'data': []
                }]
            }]
        }],
        'version': 3.2,
    }, 'Validation error'),
    ({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': '',
                    'data': []
                }]
            }]
        }],
        'version': 3.1,
    }, None),
])
async def test_edit_config_server_request(config, error, server, requests_mock):
    requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)
    requests_mock.register_uri('POST', re.compile('localhost'), real_http=True)
    requests_mock.post('http://test.iceprod/auths', status_code=200, json={})

    client = server(roles=['user'], groups=['users'])
    args = {
        'dataset_id': 'd1',
        'config': json_encode(config),
        'description': 'description',
    }
    if error:
        with pytest.raises(Exception, match=error):
            await client.request('POST', '/actions/edit_config', args)
    else:
        ret = await client.request('POST', '/actions/edit_config', args)
        assert 'result' in ret

        ret = await client.request('GET', f'/actions/edit_config/{ret["result"]}')
        assert ret['status'] == 'queued'


async def test_edit_config_server_request_bad_args(server, requests_mock):
    requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)
    requests_mock.register_uri('POST', re.compile('localhost'), real_http=True)
    requests_mock.post('http://test.iceprod/auths', status_code=200, json={})
    
    client = server(roles=['user'], groups=['users'])
    args = {}
    with pytest.raises(Exception):
        await client.request('POST', '/actions/edit_config', args)


async def test_edit_config_server_request_bad_auth(server, requests_mock):
    requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)
    requests_mock.register_uri('POST', re.compile('localhost'), real_http=True)

    requests_mock.post('http://test.iceprod/auths', status_code=403, json={})

    client = server(roles=['user'], groups=['users'])
    args = {
        'dataset_id': 'd1',
        'config': 'config',
        'description': 'description',
    }
    with pytest.raises(Exception, match='auth failed'):
        await client.request('POST', '/actions/edit_config', args)


@pytest.mark.parametrize('update,error', [
    ({'requirements': {
        'os': ['RHEL_9_x86_64'],
        'cpu': 2,
        'memory': 10.5,
        'time': 0.25,
    }}, None),
    ({'requirements': {
        'os': ['RHEL_9_x86_64'],
        'memory': '$(job)//100+2',
        'time': 0.25,
    }}, 'update requirements'),
    ({'name': 'bar'}, 'task names'),
    ({'depends': ['foo']}, 'task depends'),
])
async def test_edit_config_run(update, error, monkeypatch):
    config = {
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'data': [
                {
                    'remote': 'token:///data/sim/IceCube/2025/file.i3.zst',
                    'movement': 'input'
                }
            ]
        }],
        'version': 3.2,
    }

    monkeypatch.setattr(iceprod.services.actions.submit, 'TOKEN_PREFIXES', {
        'token://': 'https://iceprod.tokens',
    })
    token_submitter_mock = AsyncMock()
    monkeypatch.setattr(iceprod.services.actions.submit, 'TokenSubmitter', token_submitter_mock)

    queue = AsyncMock()
    api_client = AsyncMock()
    cred_client = AsyncMock()
    submit = Action(queue, logging.getLogger('action'), api_client, cred_client)

    description = 'Test dataset'

    d = Config(config)
    d.fill_defaults()
    d.validate()
    new_config = d.config
    old_config = deepcopy(new_config)
    new_config['tasks'][0].update(update)

    message = Message(
        payload={
            'dataset_id': 'd1',
            'config': json_encode(new_config),
            'description': description,
        },
        uuid=uuid4().hex,
        status='queued',
        priority=0,
        created_at=datetime.now(),
    )

    async def api(method: str, path: str, args: dict[str, Any] = {}):
        if method == 'POST' and path == f'/datasets/{message.payload["dataset_id"]}/task_actions/bulk_requirements/testing':
            return {}
        elif method == 'PUT' and path == f'/datasets/{message.payload["dataset_id"]}/description':
            return {}
        elif method == 'GET' and path == f'/datasets/{message.payload["dataset_id"]}':
            return {
                'dataset': 1,
                'dataset_id': message.payload["dataset_id"],
                'status': 'processing',
                'jobs_submitted': 10,
                'description': 'old description',
                'username': 'user',
                'group': 'users',
            }
        elif method == 'GET' and path == f'/config/{message.payload["dataset_id"]}':
            return old_config
        elif method == 'PUT' and path == f'/config/{message.payload["dataset_id"]}':
            return {}
        else:
            raise Exception('invalid path')

    api_client.request.side_effect = api
    token_submitter_mock.return_value.tokens_exist.return_value = True

    if error:
        with pytest.raises(Exception, match=error):
            await submit.run(message)
    else:
        await submit.run(message)

        api_calls = api_client.request.call_args_list
        assert len(api_calls) == 5
        assert api_calls[0][0][1] == '/datasets/d1'
        assert api_calls[1][0][1] == '/config/d1'
        assert api_calls[2][0][1] == '/datasets/d1/task_actions/bulk_requirements/testing'
        assert api_calls[3][0][1] == '/config/d1'
        assert api_calls[4][0][1] == '/datasets/d1/description'



import os
import sys
import logging
import fnmatch
import unittest
import inspect
from collections import defaultdict
from collections.abc import Iterable
from functools import wraps, partial

from tornado.concurrent import Future
from tornado.ioloop import IOLoop
from tornado.testing import gen_test

def printer(in_str,passed=True,skipped=False):
    numcols = 60
    firstpos = 0
    padding = 3 if skipped else 4
    while len(in_str) > numcols:
        # wrap longer strings
        pos = in_str.rfind(' ',firstpos,numcols)
        if pos <= firstpos or pos >= numcols:
            pos = in_str.rfind('/',firstpos,numcols)
            if pos <= firstpos or pos >= numcols:
                break
        tmp_str = in_str[0:pos+1]
        in_str = '     '+in_str[pos+1:]
        firstpos = 4
        print(tmp_str)
    # print string aligned left, and passed or failed
    final_str = in_str
    for i in range(len(in_str),numcols+padding):
        final_str += ' '

    if skipped:
        logging.error(final_str+'skipped')
        final_str += '\033[36m'+'skipped'+'\033[0m'
    elif passed:
        logging.error(final_str+'passed')
        final_str += '\033[32m'+'passed'+'\033[0m'
    else:
        logging.error(final_str+'failed')
        final_str += '\033[31m'+'failed'+'\033[0m'
    print(final_str)

def unittest_reporter(*args,**kwargs):
    def make_wrapper(obj):
        filename = obj.__globals__['__file__']
        module = os.path.basename(filename).split('_test')[0]
        prevdir = os.path.basename(os.path.dirname(filename))
        if prevdir not in ('core','server'):
            module = prevdir+'/'+module
        if 'module' in kwargs:
            module = kwargs['module']
        name = '_'.join(obj.__name__.split('_')[2:])+'()'
        if 'name' in kwargs:
            name = kwargs['name']
        if name.startswith(' '):
            test_name = '%s%s'%(module,name)
        else:
            test_name = '%s.%s'%(module,name)
        skip = False
        if 'skip' in kwargs:
            skip = kwargs['skip']
            if callable(skip):
                skip = skip()
            skip = bool(skip)
        @wraps(obj)
        def wrapper(self, *args, **kwargs):
            from iceprod.core import to_log
            if skip:
                printer('Test '+test_name,skipped=True)
                return
            try:
                with to_log(sys.stdout), to_log(sys.stderr):
                    if inspect.isgeneratorfunction(obj):
                        logging.info('test is async')
                        gen_test(obj)(self, *args,**kwargs)
                    elif inspect.iscoroutinefunction(obj):
                        logging.info('test is async')
                        IOLoop.current().run_sync(partial(obj, self, *args,**kwargs))
                    else:
                        logging.info('test is seq')
                        obj(self, *args,**kwargs)
            except Exception:
                logging.error('Error running %s test',name,
                             exc_info=True)
                printer('Test '+test_name,passed=False)
                raise
            else:
                printer('Test '+test_name)
        return wrapper
    if kwargs:
        return make_wrapper
    else:
        return make_wrapper(*args)

test_glob = '*'
def skipTest(obj, attr):
    if fnmatch.fnmatch(obj.__name__,test_glob):
        return unittest.skip()
    return lambda func: func
def glob_tests(x):
    """glob the tests that were requested"""
    return fnmatch.filter(x,test_glob)

def listmodules(package_name=''):
    """List modules in a package or directory"""
    import os
    import imp
    package_name_os = package_name.replace('.','/')
    file, pathname, description = imp.find_module(package_name_os)
    if file:
        # Not a package
        return []
    ret = []
    for module in os.listdir(pathname):
        if module.endswith('.py') and module != '__init__.py':
            tmp = os.path.splitext(module)[0]
            ret.append(package_name+'.'+tmp)
    return ret

def assertCountEqualRecursive(self, a, b, skip=[]):
    self.assertEqual(type(a), type(b))
    if isinstance(a, dict):
        if set(a).symmetric_difference(b)-set(skip):
            raise AssertionError('different keys')
        for k in a:
            if k not in skip:
                self.assertCountEqualRecursive(a[k], b[k])
    elif isinstance(a, Iterable):
        self.assertCountEqual(a, b)
    else:
        self.assertEqual(a, b)
unittest.TestCase.assertCountEqualRecursive = assertCountEqualRecursive


class services_mock(dict):
    """
    A fake `iceprod.modules.module.modules` object.

    It mocks a two-level dict of function objects,
    recording all calls and allowing different return values.
    """
    def __init__(self):
        self.called = []
        self.ret = defaultdict(dict)
    def __request(self, service, method, *args, **kwargs):
        self.called.append((service, method, args, kwargs))
        logging.info('__request: %r', self.called[-1])
        ret = self.ret[service][method]
        if isinstance(ret, Exception):
            raise ret
        f = Future()
        f.set_result(ret)
        return f
    def __contains__(self, name):
        return name in self.ret
    def __missing__(self, name):
        class Service(dict):
            def __init__(self, name, request):
                self.name = name
                self.request = request
            def __missing__(self, key):
                return partial(self.request,name,key)
        return Service(name,self.__request)

def return_once(*args, **kwargs):
    """Return every argument once, then return end_value repeatedly"""
    end_value = kwargs.pop('end_value', Exception())
    for a in args:
        yield a
    while True:
        yield end_value


def cmp_list(a,b):
    """Compare all items in a with b"""
    for aa,bb in zip(a,b):
        a_list = isinstance(aa,(list,tuple))
        b_list = isinstance(bb,(list,tuple))
        a_dict = isinstance(aa,dict)
        b_dict = isinstance(bb,dict)
        if a_list != b_list or a_dict != b_dict:
            return False
        if a_list:
            if not cmp_list(aa,bb):
                return False
        elif a_dict:
            if not cmp_dict(aa,bb):
                return False
        elif aa != bb:
            return False
    return True

def cmp_dict(a,b):
    """Compare all items in a with b"""
    for k in a:
        if k not in b:
            return False
        a_list = isinstance(a[k],(list,tuple))
        b_list = isinstance(a[k],(list,tuple))
        a_dict = isinstance(a[k],dict)
        b_dict = isinstance(b[k],dict)
        if a_list != b_list or a_dict != b_dict:
            return False
        if a_list:
            if not cmp_list(a[k],b[k]):
                return False
        elif a_dict:
            if not cmp_dict(a[k],b[k]):
                return False
        elif a[k] != b[k]:
            return False
    return True



import logging
from typing import Any

from pymongo import AsyncMongoClient
from pymongo.asynchronous.database import AsyncDatabase


type CollectionIndexes = dict[str, dict[str, Any]]
"""
Indexes for a collection.

Dict of index name : dict of keywords to create index.

Example:
    {
        "my_index": {
            "keys": [
                ("status": pymongo.ASCENDING),
                ("name": pymongo.DESCENDING)
            ]
        },
        "id_index": {
            "keys": [
                ("uuid": pymongo.ASCENDING)
            ],
            "unique": True
        }
    }
"""


class Mongo:
    def __init__(self, *, url: str, timeout: int = 60, write_concern: int = 1):
        logging_url = url.split('@')[-1] if '@' in url else url
        logging.info(f'DB: {logging_url}')
        if '/' in logging_url:
            db_url, db_name = url.rsplit('/', 1)
        else:
            db_url = url
            db_name = None

        self.client: AsyncMongoClient = AsyncMongoClient(
            db_url,
            timeoutMS=timeout*1000,
            w=write_concern,
        )
        self._db_name = db_name

    @property
    def db(self) -> AsyncDatabase:
        db_name = self._db_name
        logging.info(f'DB name: {db_name}')
        if not db_name:
            raise RuntimeError('must specify database name in __init__ url')
        return self.client[db_name]

    def __getitem__(self, name: str) -> AsyncDatabase:
        logging.info(f'DB name: {name}')
        return self.client[name]

    async def ping(self):
        await self.client.admin.command('ping')

    async def close(self):
        await self.client.close()

    async def create_indexes(self, *, db_name: str | None = None, indexes: dict[str, CollectionIndexes]):
        database: AsyncDatabase = self.client[db_name] if db_name else self.db
        for collection in indexes:
            existing = await database[collection].index_information()
            for name in indexes[collection]:
                if name not in existing:
                    logging.info('DB: creating index %s:%s', collection, name)
                    kwargs = indexes[collection][name]
                    await database[collection].create_index(name=name, **kwargs)
            for name in existing:
                if (not name.startswith('_')) and name not in indexes[collection]:
                    logging.info('DB: drop index %s:%s', collection, name)
                    await database[collection].drop_index(name)



from collections import defaultdict
import logging

from .base import authenticated, PublicHandler

logger = logging.getLogger('website-task')


class TaskBrowse(PublicHandler):
    """Handle /task urls"""
    @authenticated
    async def get(self, dataset_id):
        assert self.rest_client
        status = self.get_argument('status',default=None)
        if status:
            tasks = await self.rest_client.request('GET','/datasets/{}/tasks?status={}'.format(dataset_id,status))
            for t in tasks:
                if 'job_index' not in tasks[t]:
                    job = await self.rest_client.request('GET', '/datasets/{}/jobs/{}'.format(dataset_id, tasks[t]['job_id']))
                    tasks[t]['job_index'] = job['job_index']
            passkey = self.auth_access_token
            self.render('task_browse.html',tasks=tasks, passkey=passkey)
        else:
            status = await self.rest_client.request('GET','/datasets/{}/task_counts/status'.format(dataset_id))
            self.render('tasks.html',status=status)


class Task(PublicHandler):
    """Handle /task urls"""
    @authenticated
    async def get(self, dataset_id, task_id):
        assert self.rest_client
        status = self.get_argument('status', default=None)
        passkey = self.auth_access_token

        dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        task_details = await self.rest_client.request('GET','/datasets/{}/tasks/{}?status={}'.format(dataset_id, task_id, status))
        task_stats = await self.rest_client.request('GET','/datasets/{}/tasks/{}/task_stats?last=true'.format(dataset_id, task_id))
        if task_stats:
            task_stats = list(task_stats.values())[0]
        try:
            ret = await self.rest_client.request('GET','/datasets/{}/tasks/{}/logs?group=true'.format(dataset_id, task_id))
            logs = ret['logs']
            # logger.info("logs: %r", logs)
            ret2 = await self.rest_client.request('GET','/datasets/{}/tasks/{}/logs?keys=log_id|name|timestamp'.format(dataset_id, task_id))
            logs2 = ret2['logs']
            logger.info("logs2: %r", logs2)
            log_by_name = defaultdict(list)
            for log in sorted(logs2,key=lambda lg:lg['timestamp'] if 'timestamp' in lg else '',reverse=True):
                log_by_name[log['name']].append(log)
            for log in logs:
                if 'data' not in log or not log['data']:
                    log['data'] = ''
                log_by_name[log['name']][0] = log
        except Exception:
            log_by_name = {}
        self.render('task_detail.html', dataset=dataset, task=task_details, task_stats=task_stats, logs=log_by_name, passkey=passkey)



import asyncio
from collections.abc import Iterable
from copy import deepcopy
from dataclasses import dataclass
import importlib.resources
import json
import logging
from pathlib import Path
import re
from typing import Any
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

from cachetools.func import ttl_cache
import jsonschema
from rest_tools.client import RestClient


class ValidationError(Exception):
    pass


class ConfigSchema:
    @ttl_cache
    @staticmethod
    def list_versions() -> list[float]:
        path = Path(str(importlib.resources.files('iceprod.core') / 'data'))
        ret = []
        for p in path.glob('dataset_v*.schema.json'):
            if ver := re.match(r'dataset_v(\d\.\d).schema.json', p.name):
                ret.append(float(ver.group(1)))
        return sorted(ret)

    @ttl_cache
    @staticmethod
    def schema(version: float = 3.1) -> dict[str, Any]:
        rounded_ver = round(version, 1)
        path = importlib.resources.files('iceprod.core') / 'data' / f'dataset_v{rounded_ver}.schema.json'
        return json.loads(path.read_text())

    @ttl_cache
    @staticmethod
    def data_defaults(version: float = 3.1) -> dict[str, Any]:
        schema = ConfigSchema.schema(version)
        return {key: value.get('default', None) for key,value in schema['$defs']['data']['items']['properties'].items()}


class _ConfigMixin:
    config: dict

    def fill_defaults(self):
        """Fill in config defaults"""
        ver = self.config.get('version', None)
        if isinstance(ver, str):
            ver = float(ver)
        config_schema = ConfigSchema.schema(ver) if ver else ConfigSchema.schema()

        def _load_ref(schema_value):
            if '$ref' in list(schema_value.keys()):
                # load from ref
                parts = schema_value['$ref'].split('/')[1:]
                schema_value = config_schema
                while parts:
                    schema_value = schema_value.get(parts.pop(0), {})
                logging.debug('loading from ref: %r', schema_value)
            return schema_value

        def _fill_dict(user, schema):
            for prop in schema['properties']:
                schema_value = _load_ref(schema['properties'][prop])
                v = schema_value.get('default', None)
                if isinstance(v, (dict, list)):
                    v = deepcopy(v)  # make a copy to not use the same instance multiple times
                if prop not in user and v is not None:
                    user[prop] = v
            for k in user:
                schema_value = _load_ref(schema['properties'].get(k, {}))
                logging.debug('filling defaults for %s: %r', k, schema_value)
                try:
                    t = schema_value.get('type', 'str')
                    logging.debug('user[k] type == %r, schema_value[type] == %r', type(user[k]), t)
                    if isinstance(user[k], dict) and t == 'object':
                        _fill_dict(user[k], schema_value)
                    elif isinstance(user[k], list) and t == 'array':
                        _fill_list(user[k], schema_value)
                except KeyError:
                    logging.warning('error processing key %r with schema %r', k, schema_value)
                    raise

        def _fill_list(user, schema):
            for item in user:
                if isinstance(item, dict):
                    _fill_dict(item, schema['items'])

        _fill_dict(self.config, config_schema)

    def validate(self):
        """Validate config"""
        ver = self.config['version']
        if isinstance(ver, str):
            ver = float(ver)
            self.config['version'] = ver
        try:
            jsonschema.validate(self.config, ConfigSchema.schema(ver))
        except jsonschema.ValidationError as e:
            try:
                logging.warning("raising! %r", e.path)
                path = ''.join(f'[{p!r}]' for p in e.path)
                if isinstance(e.schema, Iterable) and 'error_msg' in e.schema:
                    msg = e.schema['error_msg']
                else:
                    msg = str(e).split('\n',1)[0]
                raise ValidationError(f'Validation error in config{path}: {msg}') from e
            except AttributeError:
                raise e


@dataclass
class Config(_ConfigMixin):
    """IceProd Dataset config"""
    config: dict


@dataclass
class Dataset(_ConfigMixin):
    """IceProd Dataset config and basic attributes"""
    dataset_id: str
    dataset_num: int
    jobs_submitted: int
    tasks_submitted: int
    tasks_per_job: int
    status: str
    priority: float
    group: str
    user: str
    debug: bool
    config: dict

    @classmethod
    async def load_from_api(cls, dataset_id: str, rest_client: RestClient) -> Self:
        dataset = await rest_client.request('GET', f'/datasets/{dataset_id}')
        config = await rest_client.request('GET', f'/config/{dataset_id}')
        return cls(
            dataset_id=dataset_id,
            dataset_num=dataset['dataset'],
            jobs_submitted=dataset['jobs_submitted'],
            tasks_submitted=dataset['tasks_submitted'],
            tasks_per_job=dataset['tasks_per_job'],
            status=dataset['status'],
            priority=dataset['priority'],
            group=dataset['group'],
            user=dataset['username'],
            debug=dataset['debug'],
            config=config,
        )


@dataclass
class Job:
    """IceProd Job instance"""
    dataset: Dataset
    job_id: str
    job_index: int
    status: str


@dataclass
class Task:
    """
    IceProd Task instance, ready for running.

    Old task stats are not loaded by default, but can be loaded on request.
    """
    dataset: Dataset
    job: Job
    task_id: str
    task_index: int
    name: str
    depends: list
    requirements: dict
    status: str
    site: str
    stats: dict
    task_files: list[dict[Any, Any]] | None = None
    instance_id: str | None = None
    oauth_tokens: list[Any] | None = None

    @classmethod
    async def load_from_api(cls, dataset_id: str, task_id: str, rest_client: RestClient) -> Self:
        d, task = await asyncio.gather(
            Dataset.load_from_api(dataset_id, rest_client),
            rest_client.request('GET', f'/datasets/{dataset_id}/tasks/{task_id}')
        )
        job = await rest_client.request('GET', f'/datasets/{dataset_id}/jobs/{task["job_id"]}')
        j = Job(d, task['job_id'], job['job_index'], job['status'])
        return cls(
            dataset=d,
            job=j,
            task_id=task['task_id'],
            task_index=task['task_index'],
            name=task['name'],
            depends=task['depends'],
            requirements=task['requirements'],
            status=task['status'],
            site=task['site'],
            stats={},
            task_files=[],
            instance_id=task.get('instance_id', '')
        )

    async def load_stats_from_api(self, rest_client: RestClient):
        ret = await rest_client.request('GET', f'/datasets/{self.dataset.dataset_id}/tasks/{self.task_id}/task_stats', {'last': 'true'})
        if not ret:
            raise Exception('No stats to load!')
        # get first (only) result in ret
        self.stats = next(iter(ret.values()))

    async def load_task_files_from_api(self, rest_client: RestClient):
        ret = await rest_client.request('GET', f'/datasets/{self.dataset.dataset_id}/files/{self.task_id}', {})
        data = []
        data_defaults = ConfigSchema.data_defaults(self.dataset.config['version'])
        for r in ret['files']:
            d = data_defaults.copy()
            d.update(r)
            data.append(d)
        self.task_files = data

    def get_task_config(self) -> dict[Any, Any]:
        return self.dataset.config['tasks'][self.task_index]



import pytest
import requests.exceptions



async def test_rest_config_err(server):
    client = server(roles=['system'])
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/config/bar')
    assert exc_info.value.response.status_code == 404


async def test_rest_config(server):
    client = server(roles=['system'])
    data = {
        'name': 'foo'
    }
    
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', '/config/bar', data)
    assert exc_info.value.response.status_code == 400
    assert 'required property' in exc_info.value.response.text

    data = {
        'tasks': [{
            'name': 'task',
            'trays': [{
                'modules': [{
                }],
            }],
        }],
    }
    await client.request('PUT', '/config/bar', data)

    ret = await client.request('GET', '/config/bar')
    assert ret == data



"""
IceProd Config Pelican Migration:
- data: all gsiftp:// moves to osdf:///
"""

import argparse
import asyncio
from collections import defaultdict
import json
import jsonschema
import logging

from pymongo import MongoClient
from rest_tools.client import RestClient
from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.core.config import Config
from iceprod.core.parser import ExpParser
from iceprod.services.actions.submit import TOKEN_PREFIXES, get_scope


notbool = lambda x: not x
strcommalist = lambda x: ','.join(x)

def gsiftp_replacement(value):
    if value.startswith('gsiftp://gridftp.icecube.wisc.edu'):
        new_value = 'osdf:///icecube/wipac' + value.split('gsiftp://gridftp.icecube.wisc.edu',1)[1]
        logging.debug('new value: %s', new_value)
        return new_value
    elif value.startswith('gsiftp://gridftp-scratch.icecube.wisc.edu'):
        #new_value = 'osdf:///icecube/wipac' + value.split('gsiftp://gridftp.icecube.wisc.edu',1)[1]
        logging.debug('old value: %s', value)
        raise Exception('scratch!')
        #return new_remote
    else:
        return value


async def cred_token(dataset_id: str, username: str, prefix: str, scope: str, token_url: str, cred_client: RestClient | None = None):
    logging.info('creating credential token for %s prefix=%s, scope=%s', dataset_id, prefix, scope)
    if username == 'rsnihur':
        username = 'i3filter'
    else:
        username = 'ice3simusr'
    if cred_client:
        # not in dry-run

        args = {
            'url': token_url,
            'transfer_prefix': prefix,
            'type': 'oauth',
        }
        ret = await cred_client.request('GET', f'/datasets/{dataset_id}/credentials', args)
        if ret and len(ret) == 1 and ret[0]['scope'] == scope:
            logging.warning('cred tokens arlready present')
        else:
            logging.warning('need to add cred tokens')
            args = {
                'username': username,
                'scope': scope,
                'url': token_url,
                'transfer_prefix': prefix,
            }
            ret = await cred_client.request('POST', '/create', args)
            await cred_client.request('DELETE', f'/datasets/{dataset_id}/credentials')
            await cred_client.request('POST', f'/datasets/{dataset_id}/credentials', ret)


async def get_tokens(config, dataset, task_files: list[dict], cred_client: RestClient | None = None):
    dataset_id = dataset['dataset_id']
    parser = ExpParser()
    username = dataset['username']
    config = config.copy()
    config['options']['jobs_submitted'] = dataset['jobs_submitted']

    # get token requests
    token_scopes = defaultdict(set)
    for task in config['tasks']:
        logging.info('get_tokens() for dataset %s task %s', dataset_id, task['name'])

        config['options']['task'] = task['name']
        task_token_scopes = defaultdict(set)
        task_data = task['data'].copy()
        for tray in task.get('trays', []):
            task_data.extend(tray.get('data',[]).copy())
            for module in tray.get('modules', []):
                task_data.extend(module.get('data',[]).copy())
        if task['task_files']:
            logging.info('adding task_files')
            task_data.extend(task_files)
        for data in task_data:
            if data['remote'] == '':
                continue
            remote = parser.parse(data['remote'], job=config)
            remote = gsiftp_replacement(remote)
            for prefix in TOKEN_PREFIXES:
                if remote.startswith(prefix):
                    if scope := get_scope(remote[len(prefix):], data['movement']):
                        logging.debug('adding scope %s for remote %s', scope, remote)
                        task_token_scopes[prefix].add(scope)
        # add in manual scopes
        for prefix,scope_str in task['token_scopes'].items():
            for scope in scope_str.split():
                if scope and '$' not in scope:
                    task_token_scopes[prefix].add(scope)
        # set token_requests
        logging.warning('scopes for task %s: %r', task['name'], dict(task_token_scopes))
        for prefix,scopes in task_token_scopes.items():
            sorted_scope_str = ' '.join(sorted(scopes))
            if '$' in sorted_scope_str:
                raise Exception('bad scope!')
            task['token_scopes'][prefix] = sorted_scope_str
            token_scopes[prefix].update(scopes)
    
    logging.warning('scopes for dataset: %r', dict(token_scopes))
    for prefix,scopes in token_scopes.items():
        sorted_scope_str = ' '.join(sorted(scopes))
        if dataset['status'] == 'processing':
            await cred_token(dataset_id, username, prefix, sorted_scope_str, TOKEN_PREFIXES[prefix], cred_client)


def convert(config):
    logging.info('Now converting config')

    if config['version'] < 3.1:
        config['version'] = 3.1

    if (steering := config.get('steering', {})) and steering.get('system', None):
        del config['steering']['system']

    for param in config.get('steering', {}).get('parameters',{}):
        value = config['steering']['parameters'][param]
        if isinstance(value, str):
            config['steering']['parameters'][param] = gsiftp_replacement(value)

    def data_cleaner(obj):
        if 'data' in obj:
            if (data := obj['data']) and isinstance(data, list):
                for d in data:
                    if remote := d.get('remote',None):
                        d['remote'] = gsiftp_replacement(remote)
                        if d['type'] != 'permanent':
                            d['type'] = 'permanent'

    for i,task in enumerate(config.get('tasks', [])):
        if 'name' not in task:
            task['name'] = f'Task{i}'
        data_cleaner(task)
        for tray in task.get('trays', []):
            data_cleaner(tray)
            for module in tray.get('modules', []):
                data_cleaner(module)

    return config


def conversion(config, output=None):
    logging.info('Converting config')
    new_config = convert(config)
    logging.info('Validating config')
    c = Config(new_config)
    c.fill_defaults()
    c.validate()
    logging.info('Config validated!')

    if output == '-':
        print(json.dumps(c.config, indent=2, sort_keys=True))
    elif output:
        with open(output, 'w') as f:
            json.dump(c.config, f, indent=2, sort_keys=True)
    return c.config


def do_mongo(server, *, dataset_id=None, min_dataset_num, max_dataset_num, output=None, cred_client=None, ignore_task_files=False, dryrun=False):

    client = MongoClient(server)
    db = client.config

    datasets = {}
    search = {'dataset': {'$gte': min_dataset_num}}
    projection = {"_id": False, "dataset_id": True, "dataset": True, "status": True, "username": True, "jobs_submitted": True}

    if dataset_id:
        search = {'dataset_id': dataset_id}

    for d in client.datasets.datasets.find(search, projection=projection):
        if (d['dataset'] >= min_dataset_num and d['dataset'] <= max_dataset_num) or dataset_id:
            datasets[d['dataset_id']] = d

    for dataset in sorted(datasets.values(), key=lambda v: v['dataset']):
        config = db.config.find_one({"dataset_id": dataset['dataset_id']}, projection={'_id': False})
        if not config:
            logging.warning('no config for dataset %s', dataset['dataset'])
            continue
        try:
            logging.warning('Processing %r', dataset['dataset'])
            new_config = conversion(config, output=output)
        except jsonschema.ValidationError:
            if dataset['status'] != 'processing' and not config['tasks'][0].get('trays'):
                logging.info('skipping dataset %r due to errors', dataset['dataset'])
                continue
            print(json.dumps(config, indent=2, sort_keys=True))
            raise
        except Exception:
            print(json.dumps(config, indent=2, sort_keys=True))
            raise
        assert config['dataset_id'] == new_config['dataset_id']

        logging.warning('Dataset %r config okay, now getting tokens', dataset['dataset'])
        search = {"dataset_id": config['dataset_id'], 'remote': {'$not': { '$type': 'object' }}}

        # exclude bad datasets in task_files
        # if config['dataset_id'] in [
        #     "4e402bd823c911eda570141877284d92",
        #     "3aaa2dfc223a11eda2f2141877284d92",
        #     "1157fac62b9d11ed8a6d141877284d92",
        #     "03f187c22a4d11ed8a6d141877284d92",
        #     "cf90ba78d77411eca198141877284d92",
        #     "02a47d3a2dfd11eda570141877284d92",
        #     "97860d2c2ba211edb99a141877284d92",
        #     "bab161fc2c0111eda570141877284d92",
        #     "aaea07461e7911edb75b141877284d92",
        #     "97cbb1ba2a6211edaa4c141877284d92",
        #     "9465e234223d11ed8a6d141877284d92",
        #     "74738da0230a11edaa4c141877284d92",
        #     "cd4b93522b9b11edb75b141877284d92",
        #     "0fa86c32204511ed8a6d141877284d92",
        #     "4123dd342b9211eda570141877284d92",
        #     "0b5387cc2d8111edb75b141877284d92",
        #     "d700e330223f11edb75b141877284d92",
        #     "701b96101f6a11edb99a141877284d92",
        #     "3aaa2dfc223a11eda2f2141877284d92",
        #     "1157fac62b9d11ed8a6d141877284d92",
        #     "4e402bd823c911eda570141877284d92",
        #     "921661582d7711eda2f2141877284d92",
        #     "6edac3440a5611ed8006141877284d92",
        #     "cf90ba78d77411eca198141877284d92",
        #     "8a5b238c2b9b11eda570141877284d92",
        #     "10b73e422bf211ed8a6d141877284d92",
        #     "03f187c22a4d11ed8a6d141877284d92",
        #     "d3a0081a230c11eda570141877284d92",
        #     "b424423a1e8411ed8a6d141877284d92",
        #     "f85e716e231611eda2f2141877284d92",
        #     "d33fcb500a5511edb80b141877284d92",
        #     "4fd54f6c242011eda570141877284d92",
        #     "5787250824c211edb99a141877284d92",
        #     "c3f29700230911edaa4c141877284d92",
        #     "ff6f79922b4c11edb75b141877284d92",
        #     "ff2a91842d9e11ed8a6d141877284d92",
        #     "3c2646aa2da411eda570141877284d92",
        #     "9d9085d81eb111eda2f2141877284d92",
        #     "51d9be4a231411edb75b141877284d92",
        #     "efb1b3fec0b211f0be81fe2e65be79b2"
        # ]:
        #     task_files = []
        # else:
        if ignore_task_files:
            task_files = []
        else:
            task_files = client.tasks.dataset_files.find(search).to_list(1000)
        asyncio.run(get_tokens(new_config, dataset, task_files, cred_client=cred_client))

        if output == '-':
            print(json.dumps(new_config, indent=2, sort_keys=True))
        elif output:
            with open(output, 'w') as f:
                json.dump(new_config, f, indent=2, sort_keys=True)

        if not dryrun:
            db.config.find_one_and_replace({'dataset_id': config['dataset_id']}, new_config)
            if task_files:
                logging.warning('updating task_files (this may take a while)')
                client.tasks.dataset_files.update_many(search, [{
                    '$set': {
                        'remote': {
                            '$replaceOne': {
                                'input': '$remote',
                                'find': 'gsiftp://gridftp.icecube.wisc.edu',
                                'replacement': 'osdf:///icecube/wipac'
                            }
                        }
                    }
                }])
        logging.warning('Completed %r', dataset['dataset'])


def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-c', '--config', help='config json file')
    group.add_argument('--mongo-server', default=None, help='mongodb server address')
    parser.add_argument('-o', '--output', default=None, help='output config json to file (or "-" for stdout)')
    parser.add_argument('--dataset-id', default=None, help='dataset id (for mongo)')
    parser.add_argument('--min-dataset-num', default=22001, type=int, help='min dataset num')
    parser.add_argument('--max-dataset-num', default=25000, type=int, help='max dataset num')
    parser.add_argument('--ignore-task-files', default=False, action='store_true', help='ignore task_files')
    parser.add_argument('--log-level', default='WARNING')
    add_auth_to_argparse(parser)
    parser.add_argument('--dry-run', action='store_true', default=False, help='do a dry run (for mongo)')
    args = parser.parse_args()

    logging.basicConfig(level=getattr(logging, args.log_level.upper()))

    if args.dry_run:
        cred_client = None
    else:
        cred_client = create_rest_client(args, retries=0, timeout=5)

    if args.config:
        logging.info('Opening config at %s', args.config)
        with open(args.config) as f:
            config = json.load(f)
        conversion(config, output=args.output)
    elif args.mongo_server:
        do_mongo(
            args.mongo_server,
            dataset_id=args.dataset_id,
            min_dataset_num=args.min_dataset_num,
            max_dataset_num=args.max_dataset_num,
            output=args.output,
            cred_client=cred_client,
            ignore_task_files=args.ignore_task_files,
            dryrun=args.dry_run
        )

if __name__ == '__main__':
    main()


"""
Test script for scheduled_tasks/dataset_completion
"""

import logging
from unittest.mock import MagicMock

import pytest
from iceprod.scheduled_tasks import dataset_completion

logger = logging.getLogger('scheduled_tasks_dataset_completion_test')


async def test_200_run():
    rc = MagicMock()
    job_counts = {}
    dataset_summaries = {'processing':['foo']}
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url.startswith('/dataset_summaries'):
            return dataset_summaries
        elif url == '/config/foo':
            return {}
        elif url == '/datasets/foo':
            return {'jobs_submitted':2, 'tasks_submitted':2}
        elif url.startswith('/datasets/foo/job_counts'):
            return job_counts
        elif url == '/datasets/foo/status' and method == 'PUT':
            client.called = True
            client.status = args['status']
            return {}
        else:
            raise Exception()
    client.called = False
    client.status = None
    rc.request = client

    logger.info('test non-buffered')
    await dataset_completion.run(rc, debug=True)
    assert not client.called

    logger.info('test processing')
    job_counts['processing'] = 2
    await dataset_completion.run(rc, debug=True)
    assert not client.called

    logger.info('test errors')
    job_counts['errors'] = 2
    del job_counts['processing']
    await dataset_completion.run(rc, debug=True)
    assert client.called
    assert client.status == 'errors'

    logger.info('test processing and suspended')
    del job_counts['errors']
    job_counts['processing'] = 1
    job_counts['suspended'] = 1
    client.called = False
    client.status = None
    await dataset_completion.run(rc, debug=True)
    assert not client.called

    logger.info('test suspended')
    del job_counts['processing']
    job_counts['suspended'] = 1
    job_counts['complete'] = 1
    await dataset_completion.run(rc, debug=True)
    assert client.called
    assert client.status == 'suspended'

    logger.info('test complete')
    del job_counts['suspended']
    job_counts['complete'] = 2
    client.called = False
    client.status = None
    await dataset_completion.run(rc, debug=True)
    assert client.called
    assert client.status == 'complete'


async def test_201_run():
    rc = MagicMock()
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url.startswith('/dataset_summaries'):
            return {'processing':['foo']}
        else:
            raise Exception()
    rc.request = client
    with pytest.raises(Exception):
        await dataset_completion.run(rc, debug=True)

    # check it normally hides the error
    await dataset_completion.run(rc, debug=False)



async def test_always_active():
    rc = MagicMock()
    dataset_summaries = {'processing':['foo']}
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url.startswith('/dataset_summaries'):
            return dataset_summaries
        elif url == '/config/foo':
            return {}
        elif url == '/datasets/foo':
            client.called = True
            return {'jobs_submitted':2, 'tasks_submitted':2, 'always_active': True}
        else:
            raise Exception()
    rc.request = client
    client.called = False

    await dataset_completion.run(rc, debug=True)
    assert client.called



from __future__ import absolute_import, division, print_function

# override multiprocessing to get coverage to work properly
import multiprocessing
class CoverageProcess(multiprocessing.Process):
    def run(self):
        import coverage
        import os
        cwd = os.getcwd()
        while os.path.basename(cwd).startswith('tmp'):
            cwd = os.path.dirname(cwd)
        cov = coverage.Coverage(data_suffix=True, branch=True,
                                source=[os.path.join(cwd,'iceprod')])
        cov.start()
        super(CoverageProcess,self).run()
        cov.stop()
        cov.save()
multiprocessing.Process = CoverageProcess

import os
import sys
import time
import signal
import logging
import logging.handlers
import argparse
import glob
import unittest

# add iceprod to PYTHONPATH
sys.path.insert(0,os.getcwd())
if 'PYTHONPATH' in os.environ:
    os.environ['PYTHONPATH'] = os.getcwd()+':'+os.environ['PYTHONPATH']
else:
    os.environ['PYTHONPATH'] = os.getcwd()

logging.basicConfig()
logger = logging.getLogger('tests')

def handler1(signum, frame):
   logger.warning('Signal handler called with signal %s' % signum)
   logger.warning('Exiting...')
   os._exit(0)

class MyArgumentParser(argparse.ArgumentParser):
    """Override exit to indicate a failure"""
    def exit(self, status=0, message=None):
        if status == 0:
            status = 2
        super(MyArgumentParser, self).exit(status=status, message=message)

parser = MyArgumentParser()
parser.add_argument('--core',action='store_true')
parser.add_argument('--server',action='store_true')
parser.add_argument('file_glob', type=str, nargs='?', default='*')
parser.add_argument('test_glob', type=str, nargs='?', default='*')
try:
    args = parser.parse_args()
except (Exception, SystemExit):
    print('exception')
    raise
if not args.core and not args.server:
    # if neither is selected, select all
    args.core = True
    args.server = True

import tests.util
tests.util.test_glob = args.test_glob

# set up logger
rootLogger = logging.getLogger()
for handler in rootLogger.handlers:
    rootLogger.removeHandler(handler)
if os.path.exists('tests.log'):
    os.remove('tests.log')
fileHandler = logging.handlers.RotatingFileHandler('tests.log','a',
                                                   67108864,1)
logformat='%(asctime)s %(levelname)s %(name)s : %(message)s'
fileHandler.setFormatter(logging.Formatter(logformat))
rootLogger.addHandler(fileHandler)
rootLogger.setLevel(logging.DEBUG)

# handle any signals
signal.signal(signal.SIGQUIT, handler1)
signal.signal(signal.SIGINT, handler1)

# start testing
logger.warning('starting...%s ' % logger.name)
print('Testing...')
print('(detailed log file available at tests.log)')
print('')

# accumulate tests
loader = unittest.defaultTestLoader
test_suites = unittest.TestSuite()
if args.core:
    test_suites.addTests(loader.discover('tests.core',args.file_glob+'_test.py'))
if args.server:
    glob_dir = '*'
    if '/' in args.file_glob:
        glob_dir = args.file_glob.split('/')[0]
        args.file_glob = args.file_glob.split('/')[-1]
    test_dirs = []
    for d in glob.glob('tests/server/'+glob_dir):
        if os.path.isdir(d) and not d.endswith('__'):
            test_dirs.append(d.replace('/','.'))
    if glob_dir == '*' or not test_dirs:
        test_dirs.append('tests.server')
    for d in test_dirs:
        print('searching dir',d,'with glob',args.file_glob)
        try:
            test_suites.addTests(loader.discover(d,args.file_glob+'_test.py'))
        except TypeError:
            logging.error(f'error searching dir {d} with glob {args.file_glob}',
                          exc_info=True)

# run tests
test_result = unittest.TestResult()
start_time = time.time()
test_suites.run(test_result)
for err in test_result.errors:
    err_str = str(err[0])+err[1]
    if any(e in err_str for e in ('ModuleImportFailure','SyntaxError','ImportError','AttributeError')):
        print(err[1])
    else:
        print(err[1])
print('-'*70)
print('Ran %d tests in %0.3fs'%(test_suites.countTestCases(),time.time()-start_time))
if not test_result.wasSuccessful():
    print('%d tests failed'%len(test_result.errors))
    sys.exit(1)



import binascii
import json
import logging
import random
from typing import Any
from unittest.mock import MagicMock

import httpx
import pytest_asyncio
from rest_tools.utils import Auth
from rest_tools.client import RestClient
import requests.exceptions
from tornado.web import create_signed_value

from iceprod.credentials.util import Client as CredClient
from iceprod.rest.auth import ROLES, GROUPS
from iceprod.website.server import Server


class ReqMock:
    def __init__(self):
        self.mocks = {}

    def add_mock(self, path: str, ret: Any) -> MagicMock:
        if callable(ret):
            m = MagicMock(side_effect=ret)
        else:
            m = MagicMock(return_value=ret)
        self.mocks[path] = m
        return m

    async def mock(self, method, path, *args, **kwargs):
        if path in self.mocks:
            return self.mocks[path](method, path, *args, **kwargs)
        else:
            response = MagicMock()
            response.status_code = 404
            raise requests.exceptions.HTTPError(response=response)


@pytest_asyncio.fixture
async def server(monkeypatch, port):
    monkeypatch.setenv('CI_TESTING', '1')
    monkeypatch.setenv('DEBUG', 'True')
    monkeypatch.setenv('PORT', str(port))

    monkeypatch.setenv('ICEPROD_CRED_ADDRESS', 'http://iceprod.test')
    monkeypatch.setenv('ICEPROD_API_ADDRESS', 'http://iceprod.test')

    token_clients = {
        'http://token.auth': {
            'client_id': 'client',
            'client_secret': 'secret',
            'transfer_prefix': ['token://'],
        }
    }
    monkeypatch.setenv('TOKEN_CLIENTS', json.dumps(token_clients))

    cookie = ''.join(hex(random.randint(0,15))[-1] for _ in range(64))
    monkeypatch.setenv('COOKIE_SECRET', cookie)
    cookie_secret = binascii.unhexlify(cookie)

    # set hostname
    hostname = b'localhost'
    address = f'http://localhost:{port}'
    monkeypatch.setenv('ICEPROD_WEB_URL', address)

    req_mock = ReqMock()
    monkeypatch.setattr(RestClient, 'request', req_mock.mock)

    def oauth_setup(self):
        self._OAUTH_AUTHORIZE_URL = 'http://idp.test/oauth/authorize'
        self._OAUTH_ACCESS_TOKEN_URL = 'http://idp.test/oauth/token'
        self._OAUTH_LOGOUT_URL = 'http://idp.test/oauth/logout'
        self._OAUTH_USERINFO_URL = 'http://idp.test/oauth/userinfo'

    auth = Auth('secret')
    monkeypatch.setattr(CredClient, 'auth', auth)

    s = Server()
    await s.start()

    def _add_to_data(data, attrs):
        logging.debug('attrs: %r', attrs)
        for item in attrs:
            key,value = item.split('=',1)
            d = data
            while '.' in key:
                k,key = key.split('.',1)
                if k not in d:
                    d[k] = {}
                d = d[k]
            if key not in d:
                d[key] = []
            d[key].append(value)

    #requests_mock.register_uri('GET', re.compile('localhost'), real_http=True)

    class Request:
        def __init__(self, token_data, timeout=None):
            self.timeout = timeout
            self.req_mock = req_mock
            self.auth = auth

            username = token_data['preferred_username']
            s.session.set(username, {'access_token': auth.create_token(username, payload=token_data)})

            self.token_cookie = create_signed_value(cookie_secret, 'iceprod_username', username)
            logging.debug('Request cookie_secret: %r', cookie_secret)

        def get_http_client(self):
            return httpx.AsyncClient(timeout=self.timeout, cookies={'iceprod_username': self.token_cookie.decode('utf-8')})

        async def request_raw(
            self,
            client: httpx.AsyncClient,
            method: str,
            path: str,
            args: dict | None = None,
            form_data: dict | None = None,
            json: dict | None = None
        ) -> httpx.Response:
            logging.debug('website request %s %s', method, path)
            kwargs = {}
            if args:
                kwargs['params'] = args
            if form_data:
                assert method != 'GET'
                assert not json
                kwargs['data'] = form_data
            if json:
                assert not form_data
                kwargs['json'] = json
            ret = await client.request(method, address+path, **kwargs)  # type: ignore
            return ret

        async def request(self, *args, **kwargs) -> str:
            async with self.get_http_client() as client:
                ret = await self.request_raw(client, *args, **kwargs)
                ret.raise_for_status()
                return ret.text


    def client(timeout=1, username='user', roles=[], groups=[], exp=10):
        data = {'preferred_username': username}
        for r in roles:
            _add_to_data(data, ROLES[r])
        for g in groups:
            _add_to_data(data, GROUPS[g])
        return Request(data, timeout=timeout)

    try:
        yield client
    finally:
        await s.stop()



from dataclasses import dataclass
import importlib.resources
import json
from hashlib import shake_128
import logging
import time

from cachetools.func import ttl_cache
import jsonschema
import jwt
from rest_tools.utils.auth import OpenIDAuth


logger = logging.getLogger('credentials.util')


CONFIG_SCHEMA = json.loads((importlib.resources.files('iceprod.credentials')/'data'/'credentials.schema.json').read_text())


@ttl_cache(maxsize=256, ttl=3600)
def _get_auth(url):
    return OpenIDAuth(url)


def get_expiration(token):
    """
    Find a token's expiration time.

    Args:
        token (str): jwt token
    Returns:
        float: expiration unix time
    """
    return jwt.decode(token, options={"verify_signature": False})['exp']


def is_expired(cred):
    """
    Check if an OAuth credential is expired.

    Will mark credential as expired if the access token has less than 5 seconds left.

    Args:
        cred (dict): credential dict
    Returns:
        bool: True if expired
    """
    if cred['type'] != 'oauth':
        return False
    return cred['expiration'] < (time.time() + 5)


@dataclass
class Client:
    url: str
    client_id: str
    client_secret: str
    transfer_prefix: list[str]

    @property
    def id(self) -> str:
        data = self.url + self.client_id
        return shake_128(data.encode('utf-8')).hexdigest(16)

    @property
    def auth(self) -> OpenIDAuth:
        return _get_auth(self.url)


class ClientCreds:
    """
    Client Credentials

    Args:
        clients: json string - dict of url to :class:Client: entries
    """
    def __init__(self, clients: str):
        self._config = json.loads(clients)

    def fill_defaults(self):
        def _fill_dict(obj, schema):
            for prop in schema.get('properties',{}):
                schema_value = schema.get('properties',{})[prop]
                v = schema_value.get('default', None)
                if prop not in obj and v is not None:
                    obj[prop] = v
            for k in obj:
                schema_value = {}
                if k in schema.get('properties',{}):
                    schema_value = schema.get('properties',{})[k]
                elif ap := schema.get('additionalProperties',{}):
                    if isinstance(ap, dict):
                        schema_value = ap
                logging.debug('filling defaults for %s: %r', k, schema_value)
                try:
                    t = schema_value.get('type', 'str')
                    logging.debug('obj[k] type == %r, schema_value[type] == %r', type(obj[k]), t)
                    if isinstance(obj[k], dict) and t == 'object':
                        _fill_dict(obj[k], schema_value)
                    elif isinstance(obj[k], list) and t == 'array':
                        _fill_list(obj[k], schema_value)
                except KeyError:
                    logging.warning('error processing key %r with schema %r', k, schema_value)
                    raise

        def _fill_list(user, schema):
            for item in user:
                if isinstance(item, dict):
                    _fill_dict(item, schema['items'])

        _fill_dict(self._config, CONFIG_SCHEMA)

    def validate(self):
        """Validate config"""
        self.fill_defaults()
        logger.debug('validating config %r against schema %r', self._config, CONFIG_SCHEMA)
        jsonschema.validate(self._config, CONFIG_SCHEMA)

    def get_auth(self, url: str) -> OpenIDAuth:
        """Get an OpenIDAuth object, to verify tokens."""
        if url not in self._config:
            raise KeyError('unknown url')
        return _get_auth(url)

    def get_client(self, url: str) -> Client:
        logger.info('getting client for %s', url)
        c = self._config[url]
        return Client(
            url=url,
            client_id=c['client_id'],
            client_secret=c['client_secret'],
            transfer_prefix=c['transfer_prefix'],
        )

    def get_clients_by_prefix(self) -> dict[str, Client]:
        ret = {}
        for url in self._config:
            c = self.get_client(url)
            for prefix in c.transfer_prefix:
                ret[prefix] = c
        return ret



"""
Detailed configuration for IceProd
"""

import importlib.resources
import json
import logging
import os
from pathlib import Path

import jsonschema  # type: ignore

logger = logging.getLogger('config')


CONFIG_SCHEMA = json.loads((importlib.resources.files('iceprod.server')/'data'/'config.schema.json').read_text())


class IceProdConfig(dict):
    """
    IceProd configuration.

    The main iceprod configuration. Designed to be modified in-program,
    not worrying about hand-editing. Currently uses a json file as backing.

    Use just like a dictionary. Note that load() and save() are called
    automatically, but are available for manual calling.

    Note that this class is not thread-safe.

    Args:
        filename: filename for config file (optional)
        defaults: use default values (optional: default True)
        validate: turn validation on/off (optional: default True)
        override: override list of key=value strings
        save: enable saving to file (optional: default True)
    """
    def __init__(self, filename=None, defaults=True, validate=True, override=None, save=True):
        if filename:
            self.filename = filename
        else:
            self.filename = os.path.expandvars('$I3PROD/etc/iceprod_config.json')

        if save and not Path(self.filename).parent.is_dir():
            raise RuntimeError('$I3PROD/etc does not exist')

        self.validate = validate
        self._save = save
        self.loading = False

        # load user input, apply defaults, and save
        self.load()
        if defaults:
            self.fill_defaults()
        self.do_validate()
        self.save()

        if override:
            self.apply_overrides(override)
            logger.info('after overrides: %s',self)

        self.do_validate()

    def fill_defaults(self):
        def _load_ref(schema_value):
            if '$ref' in list(schema_value.keys()):
                # load from ref
                parts = schema_value['$ref'].split('/')[1:]
                schema_value = CONFIG_SCHEMA
                while parts:
                    schema_value = schema_value.get(parts.pop(0), {})
                logging.debug('loading from ref: %r', schema_value)
            return schema_value

        def _fill_dict(user, schema):
            for prop in schema['properties']:
                schema_value = _load_ref(schema['properties'][prop])
                v = schema_value.get('default', None)
                if prop not in user and v is not None:
                    if isinstance(v, (dict, list)):
                        v = v.copy()
                    user[prop] = v
            for k in user:
                schema_value = _load_ref(schema['properties'].get(k, {}))
                logging.debug('filling defaults for %s: %r', k, schema_value)
                try:
                    t = schema_value.get('type', 'str')
                    logging.debug('user[k] type == %r, schema_value[type] == %r', type(user[k]), t)
                    if isinstance(user[k], dict) and t == 'object':
                        _fill_dict(user[k], schema_value)
                    elif isinstance(user[k], list) and t == 'array':
                        _fill_list(user[k], schema_value)
                except KeyError:
                    logging.warning('error processing key %r with schema %r', k, schema_value)
                    raise

        def _fill_list(user, schema):
            for item in user:
                if isinstance(item, dict):
                    _fill_dict(item, schema['items'])

        _fill_dict(self, CONFIG_SCHEMA)

    def apply_overrides(self, overrides):
        for item in overrides:
            key,val = item.split('=',1)

            # try decoding value
            if val == 'true':
                val = True
            elif val == 'false':
                val = False
            elif val.isdigit():
                val = int(val)
            else:
                try:
                    val = float(val)
                except ValueError:
                    try:
                        val = json.loads(val)
                    except Exception:
                        pass

            # put value at right key
            key = key.split('.')
            logging.debug(f'setting {key}={val}')
            obj = self
            while len(key) > 1:
                if key[0] not in obj:
                    obj[key[0]] = {}
                obj = obj[key[0]]
                key = key[1:]
            obj[key[0]] = val

    def do_validate(self):
        if self.validate:
            jsonschema.validate(self, CONFIG_SCHEMA)
        else:
            logger.warning('skipping validation of config')

    def load(self):
        """Load config from file, overwriting current contents."""
        try:
            self.loading = True
            if os.path.exists(self.filename):
                with open(self.filename) as f:
                    obj = json.load(f)
                    for key in obj:
                        self[key] = obj[key]
        except Exception:
            logger.warning('failed to load from config file %s', self.filename,
                           exc_info=True)
            raise
        finally:
            self.loading = False

    def load_string(self, text):
        """Load a config from a string, saving to file."""
        try:
            self.loading = True
            obj = json.loads(text)
            for key in obj:
                self[key] = obj[key]
            self.do_validate()
        finally:
            self.loading = False
            self.save()

    def save_to_string(self):
        return json.dumps(self, indent=4)

    def save(self):
        """Save config from file."""
        if self._save and not self.loading:
            try:
                # save securely
                with os.fdopen(os.open(self.filename+'.tmp', os.O_WRONLY | os.O_CREAT, 0o600), 'w') as f:
                    json.dump(self, f, indent=4)
                os.rename(self.filename+'.tmp', self.filename)
            except Exception:
                logger.warning('failed to save to config file %s', self.filename,
                               exc_info=True)

    # insert save function into dict methods
    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.save()

    def __delitem__(self, key):
        super().__delitem__(key)
        self.save()



import logging

import requests

from iceprod.core.jsonUtil import json_decode, json_encode
from .base import authenticated, PublicHandler

logger = logging.getLogger('website-submit')


class Config(PublicHandler):
    """Handle /config urls"""
    @authenticated
    async def get(self):
        assert self.rest_client
        dataset_id = self.get_argument('dataset_id', default=None)
        if not dataset_id:
            self.write_error(400,message='must provide dataset_id')
            return
        edit = int(self.get_query_argument('edit', default='0'))
        dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
        config = await self.rest_client.request('GET','/config/{}'.format(dataset_id))
        render_args = {
            'edit': edit,
            'dataset': dataset.get('dataset',''),
            'dataset_id': dataset_id,
            'config': json_encode(config, indent=2),
            'description': dataset.get('description',''),
            'error': '',
        }
        self.render('submit.html', **render_args)

    @authenticated
    async def post(self):
        assert self.rest_client
        dataset_id = self.get_body_argument('dataset_id', default=None)
        if not dataset_id:
            self.write_error(400,message='must provide dataset_id')
            return
        edit = int(self.get_body_argument('edit', default='0'))
        if not edit:
            self.write_error(400, message='cannot edit')
            return
        dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
        if not dataset:
            self.write_error(400, message='invalid dataset_id')
            return

        config_str = ''
        description = dataset.get('description', '')
        try:
            config_str = self.get_body_argument('submit_box')
            description = self.get_body_argument('description')
            config = json_decode(config_str)
            # if we decoded successfully, resort and reindent config
            config_str = json_encode(config, indent=2)
            config_str_compact = json_encode(config)

            args = {
                'dataset_id': dataset_id,
                'config': config_str_compact,
                'description': description,
            }
            ret = await self.rest_client.request('POST', '/actions/edit_config', args)
            id_ = ret['result']
        except requests.exceptions.HTTPError as e:
            logger.warning('failed edit', exc_info=True)
            try:
                error = e.response.json()['error']
            except Exception:
                error = e.response.text
            render_args = {
                'edit': edit,
                'dataset': dataset.get('dataset',''),
                'dataset_id': dataset_id,
                'config': config_str,
                'description': description,
                'error': error,
            }
            self.set_status(400)
            self.render('submit.html', **render_args)
        except Exception as e:
            logger.warning('failed submit', exc_info=True)
            render_args = {
                'edit': edit,
                'dataset': dataset.get('dataset',''),
                'dataset_id': dataset_id,
                'config': config_str,
                'description': description,
                'error': str(e),
            }
            self.set_status(400)
            self.render('submit.html', **render_args)
        else:
            render_args = {
                'message_id': id_,
                'status': 'queued',
                'message': '',
                'submit': False,
            }

            self.set_status(303)
            self.set_header('Location', f'/config/status/{id_}')
            self.render('submit_status.html', **render_args)


class ConfigStatus(PublicHandler):
    @authenticated
    async def get(self, id_):
        assert self.rest_client

        status = 'unknown'
        error = ''
        config_str = ''
        description = ''
        dataset = {}
        dataset_id = ''
        try:
            ret = await self.rest_client.request('GET', f'/actions/config/{id_}')
            status = ret['status']
            error = ret.get('error_message', '')
            config_str = ret['payload']['config']
            description = ret['payload']['description']
            dataset_id = ret['payload']['dataset_id']

            dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
            if not dataset:
                raise Exception('invalid dataset_id')
        except Exception as e:
            error = str(e)

        if status == 'error':
            # fix json indenting
            try:
                config = json_decode(config_str)
                config_str = json_encode(config, indent=2)
            except Exception:
                pass
            render_args = {
                'edit': '1',
                'dataset': dataset.get('dataset',''),
                'dataset_id': dataset_id,
                'config': config_str,
                'description': description,
                'error': error,
            }
            self.set_status(400)
            self.render('submit.html', **render_args)
            return
        elif status == 'complete' and dataset_id:
            self.set_status(303)
            self.set_header('Location', f'/dataset/{dataset_id}')
            error = f'Dataset updated at /dataset/{dataset_id}'

        render_args = {
            'message_id': id_,
            'status': status,
            'message': error,
            'submit': False,
        }

        self.render('submit_status.html', **render_args)


DEFAULT_CONFIG = {
    "categories": [],
    "dataset": 0,
    "description": "",
    "difplus": None,
    "options": {},
    "parent_id": 0,
    "steering": None,
    "tasks": [],
    "version": 3.2
}


class Submit(PublicHandler):
    """Handle /submit urls"""
    @authenticated
    async def get(self):
        config = json_encode(DEFAULT_CONFIG.copy(), indent=2)

        error = ''
        if e := self.get_argument('error', None):
            error = e

        render_args = {
            'edit': False,
            'dataset': '',
            'dataset_id': '',
            'config': config,
            'groups': self.auth_groups,
            'group': '',
            'description': '',
            'jobs_submitted': 1,
            'error': error,
        }
        self.render('submit.html', **render_args)

    @authenticated
    async def post(self):
        assert self.rest_client
        logger.info('new dataset submission!')

        config_str = ''
        description = ''
        njobs = 1
        group = ''
        try:
            config_str = self.get_body_argument('submit_box')
            description = self.get_body_argument('description')
            group = self.get_body_argument('group')
            njobs = int(self.get_body_argument('number_jobs'))
            config = json_decode(config_str)
            # if we decoded successfully, resort and reindent config
            config_str = json_encode(config, indent=2)
            config_str_compact = json_encode(config)

            args = {
                'config': config_str_compact,
                'description': description,
                'jobs_submitted': njobs,
                'username': self.current_user,
                'group': group,
            }
            ret = await self.rest_client.request('POST', '/actions/submit', args)
            id_ = ret['result']
        except requests.exceptions.HTTPError as e:
            logger.warning('failed submit', exc_info=True)
            try:
                error = e.response.json()['error']
            except Exception:
                error = e.response.text
            render_args = {
                'passkey': '',
                'edit': False,
                'dataset': '',
                'dataset_id': '',
                'config': config_str,
                'groups': self.auth_groups,
                'group': group,
                'description': description,
                'jobs_submitted': njobs,
                'error': error,
            }
            self.set_status(400)
            self.render('submit.html', **render_args)
        except Exception as e:
            logger.warning('failed submit', exc_info=True)
            render_args = {
                'passkey': '',
                'edit': False,
                'dataset': '',
                'dataset_id': '',
                'config': config_str,
                'groups': self.auth_groups,
                'group': group,
                'description': description,
                'jobs_submitted': njobs,
                'error': str(e),
            }
            self.set_status(400)
            self.render('submit.html', **render_args)
        else:
            render_args = {
                'message_id': id_,
                'status': 'queued',
                'message': '',
                'submit': True,
            }

            self.set_status(303)
            self.set_header('Location', f'/submit/status/{id_}')
            self.render('submit_status.html', **render_args)


class SubmitStatus(PublicHandler):
    @authenticated
    async def get(self, id_):
        assert self.rest_client

        status = 'unknown'
        error = ''
        config = ''
        description = ''
        jobs_submitted = 1
        group = ''
        dataset_id = None
        try:
            ret = await self.rest_client.request('GET', f'/actions/submit/{id_}')
            status = ret['status']
            error = ret.get('error_message', '')
            config = ret['payload']['config']
            description = ret['payload']['description']
            jobs_submitted = int(ret['payload']['jobs_submitted'])
            group = ret['payload']['group']
            dataset_id = ret['payload'].get('dataset_id', '')
        except Exception as e:
            error = str(e)

        if status == 'error':
            render_args = {
                'passkey': '',
                'edit': False,
                'dataset': '',
                'dataset_id': '',
                'config': config,
                'groups': self.auth_groups,
                'group': group,
                'description': description,
                'jobs_submitted': jobs_submitted,
                'error': error,
            }
            self.render('submit.html', **render_args)
            return
        elif status == 'complete' and dataset_id:
            self.set_status(303)
            self.set_header('Location', f'/dataset/{dataset_id}')
            error = f'Dataset created at /dataset/{dataset_id}'

        render_args = {
            'message_id': id_,
            'status': status,
            'message': error,
            'submit': True,
        }

        self.render('submit_status.html', **render_args)
